# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence
import dataclasses
import typing

from ansys.api.acp.v0 import oriented_selection_set_pb2, oriented_selection_set_pb2_grpc

from .._utils.array_conversions import to_1D_double_array, to_tuple_from_1D_array
from .._utils.property_protocols import ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.linked_object_list import (
    define_linked_object_list,
    define_polymorphic_linked_object_list,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._mesh_data import full_mesh_property, shell_mesh_property
from .base import CreatableTreeObject, IdTreeObject
from .boolean_selection_rule import BooleanSelectionRule
from .cylindrical_selection_rule import CylindricalSelectionRule
from .element_set import ElementSet
from .enums import (
    DrapingMaterialModel,
    RosetteSelectionMethod,
    draping_material_type_from_pb,
    draping_material_type_to_pb,
    rosette_selection_method_from_pb,
    rosette_selection_method_to_pb,
    status_type_from_pb,
)
from .lookup_table_1d_column import LookUpTable1DColumn
from .lookup_table_3d_column import LookUpTable3DColumn
from .object_registry import register
from .parallel_selection_rule import ParallelSelectionRule
from .rosette import Rosette
from .spherical_selection_rule import SphericalSelectionRule
from .tube_selection_rule import TubeSelectionRule
from .variable_offset_selection_rule import VariableOffsetSelectionRule

__all__ = [
    "OrientedSelectionSet",
    "OrientedSelectionSetElementalData",
    "OrientedSelectionSetNodalData",
]

if typing.TYPE_CHECKING:  # pragma: no cover
    # Since the 'LinkedSelectionRule' class is used by the boolean selection rule,
    # this would cause a circular import at run-time.
    from .. import BooleanSelectionRule, GeometricalSelectionRule

_SELECTION_RULES_LINKABLE_TO_OSS: typing.TypeAlias = typing.Union[
    ParallelSelectionRule,
    CylindricalSelectionRule,
    SphericalSelectionRule,
    TubeSelectionRule,
    "GeometricalSelectionRule",
    VariableOffsetSelectionRule,
    "BooleanSelectionRule",
]


@dataclasses.dataclass
class OrientedSelectionSetElementalData(ElementalData):
    """Represents elemental data for an Oriented Selection Set."""

    normal: VectorData | None = None
    orientation: VectorData | None = None
    reference_direction: VectorData | None = None


@dataclasses.dataclass
class OrientedSelectionSetNodalData(NodalData):
    """Represents nodal data for an Oriented Selection Set."""


@mark_grpc_properties
@register
class OrientedSelectionSet(CreatableTreeObject, IdTreeObject):
    """Instantiate an Oriented Selection Set.

    Parameters
    ----------
    name :
        The name of the Oriented Selection Set.
    element_sets :
        Element Sets on which the Oriented Selection Set is defined.
    orientation_point :
        The Orientation Point of the Oriented Selection Set.
    orientation_direction :
        The Orientation Direction of the Oriented Element set.
    rosettes :
        Rosettes of the Oriented Selection Set.
    rosette_selection_method :
        Selection Method for Rosettes of the Oriented Selection Set.
    selection_rules :
        Selection Rules which may limit the extent of the Oriented Selection Set.
    draping :
        If ``True``, activate draping to adjust the reference directions.
    draping_seed_point :
        Starting point of the draping algorithm.
    auto_draping_direction :
        If ``True``, the reference direction of the Oriented Selection Set at
        the seed point is used as draping direction.
    draping_direction :
        Set the primary draping direction for the draping algorithm. Only used if
        ``auto_draping_direction`` is ``False``.
    draping_mesh_size :
        Defines the mesh size for the draping algorithm. If set to ``-1.``, the
        mesh size is automatically determined based on the average element size.
    draping_material_model :
        Chooses between different draping formulations.
    draping_ud_coefficient :
        Value between ``0`` and ``1`` which determines the amount of deformation
        in the transverse direction if the draping material model is set to
        :attr:`DrapingMaterialModel.UD`.
    rotation_angle :
        Angle in degrees by which the initial reference directions are rotated around the orientations.
    reference_direction_field :
        A 3D lookup table column that defines the reference directions.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "oriented_selection_sets"
    _OBJECT_INFO_TYPE = oriented_selection_set_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = oriented_selection_set_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "OrientedSelectionSet",
        element_sets: Sequence[ElementSet] = tuple(),
        orientation_point: tuple[float, float, float] = (0.0, 0.0, 0.0),
        orientation_direction: tuple[float, float, float] = (0.0, 0.0, 0.0),
        rosettes: Sequence[Rosette] = tuple(),
        rosette_selection_method: RosetteSelectionMethod = "minimum_angle",
        selection_rules: Sequence[_SELECTION_RULES_LINKABLE_TO_OSS] = tuple(),
        draping: bool = False,
        draping_seed_point: tuple[float, float, float] = (0.0, 0.0, 0.0),
        auto_draping_direction: bool = True,
        draping_direction: tuple[float, float, float] = (0.0, 0.0, 1.0),
        use_default_draping_mesh_size: bool = True,
        draping_mesh_size: float = 0.0,
        draping_material_model: DrapingMaterialModel = DrapingMaterialModel.WOVEN,
        draping_ud_coefficient: float = 0.0,
        rotation_angle: float = 0.0,
        reference_direction_field: LookUpTable1DColumn | LookUpTable3DColumn | None = None,
    ):
        super().__init__(name=name)
        self.element_sets = element_sets
        self.orientation_point = orientation_point
        self.orientation_direction = orientation_direction
        self.rosettes = rosettes
        self.rosette_selection_method = RosetteSelectionMethod(rosette_selection_method)
        self.selection_rules = selection_rules
        self.draping = draping
        self.draping_seed_point = draping_seed_point
        self.auto_draping_direction = auto_draping_direction
        self.draping_direction = draping_direction
        self.use_default_draping_mesh_size = use_default_draping_mesh_size
        self.draping_mesh_size = draping_mesh_size
        self.draping_material_model = DrapingMaterialModel(draping_material_model)
        self.draping_ud_coefficient = draping_ud_coefficient
        self.rotation_angle = rotation_angle
        self.reference_direction_field = reference_direction_field

    def _create_stub(self) -> oriented_selection_set_pb2_grpc.ObjectServiceStub:
        return oriented_selection_set_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    element_sets = define_linked_object_list("properties.element_sets", ElementSet)

    orientation_point = grpc_data_property(
        "properties.orientation_point",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    orientation_direction = grpc_data_property(
        "properties.orientation_direction",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )

    rosettes = define_linked_object_list("properties.rosettes", Rosette)
    rosette_selection_method = grpc_data_property(
        "properties.rosette_selection_method",
        from_protobuf=rosette_selection_method_from_pb,
        to_protobuf=rosette_selection_method_to_pb,
    )

    draping: ReadWriteProperty[bool, bool] = grpc_data_property("properties.draping")
    draping_seed_point = grpc_data_property(
        "properties.draping_seed_point",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    auto_draping_direction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.auto_draping_direction"
    )
    draping_direction = grpc_data_property(
        "properties.draping_direction",
        from_protobuf=to_tuple_from_1D_array,
        to_protobuf=to_1D_double_array,
    )
    use_default_draping_mesh_size: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_draping_mesh_size"
    )
    draping_mesh_size: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.draping_mesh_size"
    )
    draping_material_model = grpc_data_property(
        "properties.draping_material_model",
        from_protobuf=draping_material_type_from_pb,
        to_protobuf=draping_material_type_to_pb,
    )
    draping_ud_coefficient: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.draping_ud_coefficient"
    )

    rotation_angle: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.rotation_angle"
    )

    selection_rules = define_polymorphic_linked_object_list(
        "properties.selection_rules",
        allowed_types=(
            ParallelSelectionRule,
            CylindricalSelectionRule,
            SphericalSelectionRule,
            TubeSelectionRule,
            VariableOffsetSelectionRule,
            BooleanSelectionRule,
        ),
    )

    reference_direction_field = grpc_link_property(
        "properties.reference_direction_field",
        allowed_types=(LookUpTable3DColumn, LookUpTable1DColumn),
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property

    elemental_data = elemental_data_property(OrientedSelectionSetElementalData)
    nodal_data = nodal_data_property(OrientedSelectionSetNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable, Sequence
import dataclasses
import typing
from typing import Any, cast

import numpy as np
from packaging.version import Version
from packaging.version import parse as parse_version

from ansys.api.acp.v0 import (
    boolean_selection_rule_pb2_grpc,
    cad_geometry_pb2_grpc,
    cutoff_selection_rule_pb2_grpc,
    cylindrical_selection_rule_pb2_grpc,
    edge_set_pb2_grpc,
    element_set_pb2_grpc,
    enum_types_pb2,
    fabric_pb2_grpc,
    field_definition_pb2_grpc,
    geometrical_selection_rule_pb2_grpc,
    imported_modeling_group_pb2_grpc,
    imported_solid_model_pb2_grpc,
    lookup_table_1d_pb2_grpc,
    lookup_table_3d_pb2_grpc,
    material_pb2,
    material_pb2_grpc,
    model_pb2,
    model_pb2_grpc,
    modeling_group_pb2_grpc,
    modeling_ply_pb2_grpc,
    oriented_selection_set_pb2_grpc,
    parallel_selection_rule_pb2_grpc,
    ply_geometry_export_pb2,
    rosette_pb2_grpc,
    sampling_point_pb2_grpc,
    section_cut_pb2_grpc,
    sensor_pb2_grpc,
    solid_model_pb2_grpc,
    spherical_selection_rule_pb2_grpc,
    stackup_pb2_grpc,
    sublaminate_pb2_grpc,
    tube_selection_rule_pb2_grpc,
    variable_offset_selection_rule_pb2_grpc,
    virtual_geometry_pb2_grpc,
)
from ansys.api.acp.v0.base_pb2 import CollectionPath

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from .._utils.resource_paths import join as rp_join
from .._utils.typing_helper import PATH as _PATH
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    ScalarData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.enum_wrapper import wrap_to_string_enum
from ._grpc_helpers.exceptions import wrap_grpc_errors
from ._grpc_helpers.mapping import define_create_method, define_mutable_mapping
from ._grpc_helpers.property_helper import (
    _PROTOBUF_T,
    _set_data_attribute,
    grpc_data_property,
    grpc_data_property_read_only,
    mark_grpc_properties,
)
from ._grpc_helpers.protocols import ObjectInfo
from ._grpc_helpers.supported_since import supported_since
from ._mesh_data import full_mesh_property, shell_mesh_property, solid_mesh_property
from .base import ServerWrapper, TreeObject
from .boolean_selection_rule import BooleanSelectionRule
from .cad_geometry import CADGeometry
from .cut_off_selection_rule import CutOffSelectionRule
from .cylindrical_selection_rule import CylindricalSelectionRule
from .edge_set import EdgeSet
from .element_set import ElementSet
from .enums import (
    ArrowType,
    OffsetType,
    PlyGeometryExportFormat,
    UnitSystemType,
    arrow_type_to_pb,
    offset_type_to_pb,
    ply_geometry_export_format_to_pb,
    unit_system_type_from_pb,
    unit_system_type_to_pb,
)
from .fabric import Fabric
from .field_definition import FieldDefinition
from .geometrical_selection_rule import GeometricalSelectionRule
from .imported_modeling_group import ImportedModelingGroup
from .imported_solid_model import ImportedSolidModel
from .lookup_table_1d import LookUpTable1D
from .lookup_table_3d import LookUpTable3D
from .material import Material
from .modeling_group import ModelingGroup
from .modeling_ply import ModelingPly
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet
from .parallel_selection_rule import ParallelSelectionRule
from .rosette import Rosette
from .sampling_point import SamplingPoint
from .section_cut import SectionCut
from .sensor import Sensor
from .solid_model import SolidModel
from .spherical_selection_rule import SphericalSelectionRule
from .stackup import Stackup
from .sublaminate import SubLaminate
from .tube_selection_rule import TubeSelectionRule
from .utils import CoordinateTransformation
from .variable_offset_selection_rule import VariableOffsetSelectionRule
from .virtual_geometry import VirtualGeometry

__all__ = [
    "FeFormat",
    "HDF5CompositeCAEImportMode",
    "HDF5CompositeCAEProjectionMode",
    "IgnorableEntity",
    "Model",
    "ModelElementalData",
    "ModelNodalData",
    "ShellMappingProperties",
    "SolidMappingProperties",
]

FeFormat, fe_format_to_pb, _ = wrap_to_string_enum(
    "FeFormat",
    enum_types_pb2.FileFormat,
    module=__name__,
    value_converter=lambda val: val.lower().replace("_", ":"),
    doc="Options for the format of the FE file.",
    explicit_value_list=(
        enum_types_pb2.FileFormat.ANSYS_H5,
        enum_types_pb2.FileFormat.ANSYS_CDB,
        enum_types_pb2.FileFormat.ANSYS_DAT,
    ),
)
IgnorableEntity, ignorable_entity_to_pb, _ = wrap_to_string_enum(
    "IgnorableEntity",
    model_pb2.LoadFromFEFileRequest.IgnorableEntity,
    module=__name__,
    doc="Options for the entities to ignore when loading an FE file.",
)
HDF5CompositeCAEImportMode, hdf5_composite_cae_import_mode_to_pb, _ = wrap_to_string_enum(
    "HDF5CompositeCAEImportMode",
    model_pb2.ImportMode,
    module=__name__,
    doc="Options for the import mode of the HDF5 Composite CAE file.",
)
HDF5CompositeCAEProjectionMode, hdf5_composite_cae_projection_mode_to_pb, _ = wrap_to_string_enum(
    "HDF5CompositeCAEProjectionMode",
    model_pb2.ProjectionMode,
    module=__name__,
    doc="Options for the projection mode of the HDF5 Composite CAE file.",
)


@dataclasses.dataclass
class ModelElementalData(ElementalData):
    """Represents elemental data for a Model."""

    normal: VectorData | None = None
    thickness: ScalarData[np.float64] | None = None
    relative_thickness_correction: ScalarData[np.float64] | None = None
    area: ScalarData[np.float64] | None = None
    price: ScalarData[np.float64] | None = None
    volume: ScalarData[np.float64] | None = None
    mass: ScalarData[np.float64] | None = None
    offset: ScalarData[np.float64] | None = None
    cog: VectorData | None = None

    @classmethod
    def _field_names(cls, server_version: Version | None = None) -> list[str]:
        """Bugfix override for #717.

        Override the base class _field_names method to remove the 'price' field
        if the server version is less than 25.2.

        Before 25.2, retrieving the 'price' could crash the server if the model
        contains void analysis plies (on an imported solid model).
        In this way, the other properties can still be accessed, and we can avoid
        the crash.
        """
        res = super()._field_names(server_version)
        if server_version is not None and server_version < parse_version("25.2"):
            res.remove("price")
        return res


@dataclasses.dataclass
class ModelNodalData(NodalData):
    """Represents nodal data for a Model."""


@dataclasses.dataclass
class ShellMappingProperties:
    """Properties for mapping to the shell on importing HDF5 Composite CAE files."""

    all_elements: bool = True
    element_sets: Sequence[ElementSet] = ()
    relative_thickness_tolerance: float = 0.5
    relative_in_plane_tolerance: float = 0.01
    angle_tolerance: float = 35.0
    small_hole_threshold: float = 0.0


@dataclasses.dataclass
class SolidMappingProperties:
    """Properties for importing HDF5 Composite CAE files as imported plies."""

    offset_type: OffsetType = OffsetType.BOTTOM_OFFSET


@mark_grpc_properties
@register
class Model(TreeObject):
    """Defines an ACP Model.

    Wrapper for accessing an ACP Model residing on a server.

    Parameters
    ----------
    name :
        The name of the model.
    use_nodal_thicknesses :
        Defines whether to use nodal thicknesses or element thicknesses.
    draping_offset_correction :
        Defines whether to consider lay-up thickness in draping analysis.
    angle_tolerance :
        Section computation angle tolerance (in degree).
    relative_thickness_tolerance :
        Section computation relative thickness tolerance.
    minimum_analysis_ply_thickness :
        Section computation minimum analysis ply thickness (in length
        unit of model).
    force_disable_result_extrapolation :
        Force the result extrapolation to be disabled ('ERESX,NO' command)
        when exporting to APDL or CDB format. Available since ACP server
        version 25.2.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "models"
    _OBJECT_INFO_TYPE = model_pb2.ObjectInfo
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "ACP Model",
        use_nodal_thicknesses: bool = False,
        draping_offset_correction: bool = False,
        angle_tolerance: float = 1.0,
        relative_thickness_tolerance: float = 0.01,
        minimum_analysis_ply_thickness: float = 1e-6,
        force_disable_result_extrapolation: bool = False,
    ) -> None:
        super().__init__(name=name)

        self.use_nodal_thicknesses = use_nodal_thicknesses
        self.draping_offset_correction = draping_offset_correction
        self.angle_tolerance = angle_tolerance
        self.relative_thickness_tolerance = relative_thickness_tolerance
        self.minimum_analysis_ply_thickness = minimum_analysis_ply_thickness
        self.force_disable_result_extrapolation = force_disable_result_extrapolation

    def _get_stub(self) -> model_pb2_grpc.ObjectServiceStub:
        return cast(model_pb2_grpc.ObjectServiceStub, super()._get_stub())

    def _create_stub(self) -> model_pb2_grpc.ObjectServiceStub:
        return model_pb2_grpc.ObjectServiceStub(self._channel)

    # # TODO: document further properties, or autogenerate docstring from .proto files.

    use_nodal_thicknesses: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_nodal_thicknesses"
    )
    draping_offset_correction: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.draping_offset_correction"
    )
    angle_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.angle_tolerance"
    )
    relative_thickness_tolerance: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.relative_thickness_tolerance"
    )
    minimum_analysis_ply_thickness: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.minimum_analysis_ply_thickness"
    )
    force_disable_result_extrapolation: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.force_disable_result_extrapolation",
        readable_since="25.2",
        writable_since="25.2",
    )

    @staticmethod
    def _set_unit_system_data_attribute(pb_obj: ObjectInfo, name: str, value: _PROTOBUF_T) -> None:
        # remove the 'minimum_analysis_ply_thickness' property from the pb object, to
        # allow the backend to convert it to the new unit system.
        pb_obj.properties.ClearField("minimum_analysis_ply_thickness")
        _set_data_attribute(pb_obj, name, value)

    unit_system = grpc_data_property(
        "properties.unit_system",
        from_protobuf=unit_system_type_from_pb,
        to_protobuf=unit_system_type_to_pb,
        setter_func=_set_unit_system_data_attribute,
        writable_since="25.1",
    )

    average_element_size: ReadOnlyProperty[float] = grpc_data_property_read_only(
        "properties.average_element_size"
    )

    @classmethod
    def _from_file(cls, *, path: _PATH, server_wrapper: ServerWrapper) -> Model:
        """Instantiate a Model from an ACPH5 file.

        Parameters
        ----------
        path:
            File path, on the server.
        server_wrapper:
            Representation of the ACP instance.
        """
        request = model_pb2.LoadFromFileRequest(path=server_wrapper.auto_upload(path))
        with wrap_grpc_errors():
            reply = model_pb2_grpc.ObjectServiceStub(server_wrapper.channel).LoadFromFile(request)
        return cls._from_object_info(object_info=reply, server_wrapper=server_wrapper)

    @classmethod
    def _from_fe_file(
        cls,
        *,
        path: _PATH,
        server_wrapper: ServerWrapper,
        format: FeFormat,  # type: ignore
        ignored_entities: Iterable[IgnorableEntity] = (),  # type: ignore
        convert_section_data: bool = False,
        unit_system: UnitSystemType = UnitSystemType.FROM_FILE,
    ) -> Model:
        """Load the model from an FE file.

        Parameters
        ----------
        path:
            File path, on the server.
        server_wrapper:
            Representation of the ACP instance.
        format:
            Format of the FE file. Can be one of ``"ansys:h5"``, ``"ansys:cdb"``,
            ``"ansys:dat"``, ``"abaqus:inp"``, or ``"nastran:bdf"``.
        ignored_entities:
            Entities to ignore when loading the FE file. Can be a subset of
            the following values:
            ``"coordinate_systems"``, ``"element_sets"``, ``"materials"``,
            ``"mesh"``, or ``"shell_sections"``.
        convert_section_data:
            Whether to import the section data of a shell model and convert it
            into ACP composite definitions.
        unit_system:
            Defines the unit system of the imported file. Must be set if the
            input file does not have units. If the input file does have units,
            ``unit_system`` must be either ``"from_file"``, or match the input
            unit system.
        """
        format_pb = fe_format_to_pb(format)
        ignored_entities_pb = [ignorable_entity_to_pb(val) for val in ignored_entities]

        request = model_pb2.LoadFromFEFileRequest(
            path=server_wrapper.auto_upload(path),
            format=cast(Any, format_pb),
            ignored_entities=cast(Any, ignored_entities_pb),
            convert_section_data=convert_section_data,
            unit_system=cast(Any, unit_system_type_to_pb(unit_system)),
        )
        with wrap_grpc_errors():
            reply = model_pb2_grpc.ObjectServiceStub(server_wrapper.channel).LoadFromFEFile(request)
        return cls._from_object_info(object_info=reply, server_wrapper=server_wrapper)

    def update(self, *, relations_only: bool = False) -> None:
        """Update the model.

        Parameters
        ----------
        relations_only :
            Whether to update and propagate only the status of all objects.
        """
        with wrap_grpc_errors():
            self._get_stub().Update(
                model_pb2.UpdateRequest(
                    resource_path=self._resource_path, relations_only=relations_only
                )
            )

    def save(self, path: _PATH, *, save_cache: bool = True) -> None:
        """
        Save ACP Model (.acph5).

        Parameters
        ----------
        path:
            File path.
        save_cache:
            Whether to store the update results such as Analysis Plies and solid models.
        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().SaveToFile(
                    model_pb2.SaveToFileRequest(
                        resource_path=self._resource_path,
                        path=export_path,
                        save_cache=save_cache,
                    )
                )

    def export_analysis_model(self, path: _PATH) -> None:
        """Save the analysis model to a CDB file.

        Parameters
        ----------
        path:
            Target file path. E.g. /tmp/ACPAnalysisModel.cdb
        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().SaveAnalysisModel(
                    model_pb2.SaveAnalysisModelRequest(
                        resource_path=self._resource_path,
                        path=export_path,
                    )
                )

    @supported_since("25.1")
    def export_hdf5_composite_cae(
        self,
        path: _PATH,
        *,
        remove_midside_nodes: bool = True,
        layup_representation_3d: bool = False,
        offset_type: OffsetType = OffsetType.BOTTOM_OFFSET,
        all_elements: bool = True,
        element_sets: Sequence[ElementSet | OrientedSelectionSet] = (),
        all_plies: bool = True,
        plies: Sequence[ModelingGroup | ModelingPly] = (),
        ascii_encoding: bool = False,
    ) -> None:
        """
        Export the lay-up to the HDF5 Composite CAE format.

        Parameters
        ----------
        path :
            File path.
        remove_midside_nodes :
            If True, remove mid-side nodes from the exported mesh. This increases the
            overall performance.
        layup_representation_3d :
            If True, the 3D representation of the lay-up is computed, and the offset ply
            surfaces are exported.
        offset_type :
            Defines if the bottom, mid, or top surface of the plies is exported.
            Only used if ``layup_representation_3d`` is True.
        all_elements :
            Whether to limit the export to some user-defined element sets or not.
        element_sets :
            Only plies defined on the selected element sets or oriented selection
            sets will be exported.
            Used only if ``all_elements`` is False.
        all_plies :
            Whether to export all plies or a user-defined set.
        plies :
            User-defined set of Modeling Plies and/or Modeling Groups.
            Used only if ``all_plies`` is False.
        ascii_encoding :
            If True, use ASCII encoding when writing text attributes to the HDF5 CAE
            file. This may be needed for compatibility with programs that don't fully
            support unicode when reading the file.
        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().ExportHDF5CompositeCAE(
                    model_pb2.ExportHDF5CompositeCAERequest(
                        resource_path=self._resource_path,
                        path=export_path,
                        remove_midside_nodes=remove_midside_nodes,
                        layup_representation_3d=layup_representation_3d,
                        offset_type=offset_type_to_pb(offset_type),  # type: ignore
                        ascii_encoding=ascii_encoding,
                        all_elements=all_elements,
                        element_sets=[element_set._resource_path for element_set in element_sets],
                        all_plies=all_plies,
                        plies=[ply._resource_path for ply in plies],
                    )
                )

    @supported_since("25.1")
    def import_hdf5_composite_cae(
        self,
        path: _PATH,
        import_mode: HDF5CompositeCAEImportMode = HDF5CompositeCAEImportMode.APPEND,  # type: ignore
        projection_mode: HDF5CompositeCAEProjectionMode = HDF5CompositeCAEProjectionMode.SHELL,  # type: ignore
        minimum_angle_tolerance: float = 0.001,
        recompute_reference_directions: bool = False,
        shell_mapping_properties: ShellMappingProperties = ShellMappingProperties(),
        solid_mapping_properties: SolidMappingProperties = SolidMappingProperties(),
        coordinate_transformation: CoordinateTransformation = CoordinateTransformation(),
    ) -> None:
        """Import the lay-up from an HDF5 Composite CAE file.

        Parameters
        ----------
        path :
            File path.
        import_mode :
            In :py:attr:`.HDF5CompositeCAEImportMode.APPEND` mode, the imported objects are
            appended to existing layup.
            In :py:attr:`.HDF5CompositeCAEImportMode.OVERWRITE` mode, existing objects in the model
            with the same name are replaced if possible (not locked).
        projection_mode :
            Determines whether loaded plies are mapped onto the reference surface
            (:py:attr:`.HDF5CompositeCAEProjectionMode.SHELL` mode) or exposed as
            3D plies (:py:attr:`.HDF5CompositeCAEProjectionMode.SOLID` mode).
        minimum_angle_tolerance :
            Minimum angle tolerance for which tabular correction angles for plies are computed.
        recompute_reference_directions :
            Whether reference directions should be recomputed from tabular angle data or not.
        shell_mapping_properties :
            Properties for mapping to the shell on importing HDF5 Composite CAE files.
            Used only if ``projection_mode`` is set to ``"shell"``.
        solid_mapping_properties :
            Properties for importing HDF5 Composite CAE files as imported plies.
            Used only if ``projection_mode`` is set to ``"solid"``.
        coordinate_transformation :
            Coordinate transformation applied to the imported lay-up.
        """
        if projection_mode == HDF5CompositeCAEProjectionMode.SHELL:
            mapping_properties_kwargs: (
                dict[str, model_pb2.ShellMappingProperties]
                | dict[str, model_pb2.SolidMappingProperties]
            ) = {
                "shell_mapping_properties": model_pb2.ShellMappingProperties(
                    all_elements=shell_mapping_properties.all_elements,
                    element_sets=[
                        element_set._resource_path
                        for element_set in shell_mapping_properties.element_sets
                    ],
                    relative_thickness_tolerance=shell_mapping_properties.relative_thickness_tolerance,
                    relative_in_plane_tolerance=shell_mapping_properties.relative_in_plane_tolerance,
                    angle_tolerance=shell_mapping_properties.angle_tolerance,
                    small_hole_threshold=shell_mapping_properties.small_hole_threshold,
                ),
            }
        else:
            assert projection_mode == HDF5CompositeCAEProjectionMode.SOLID
            mapping_properties_kwargs = {
                "solid_mapping_properties": model_pb2.SolidMappingProperties(
                    offset_type=offset_type_to_pb(solid_mapping_properties.offset_type)  # type: ignore
                ),
            }

        with wrap_grpc_errors():
            self._get_stub().ImportHDF5CompositeCAE(
                model_pb2.ImportHDF5CompositeCAERequest(
                    resource_path=self._resource_path,
                    path=self._server_wrapper.auto_upload(path),
                    import_mode=hdf5_composite_cae_import_mode_to_pb(import_mode),  # type: ignore
                    projection_mode=hdf5_composite_cae_projection_mode_to_pb(projection_mode),  # type: ignore
                    minimum_angle_tolerance=minimum_angle_tolerance,
                    recompute_reference_directions=recompute_reference_directions,
                    coordinate_transformation=model_pb2.CoordinateTransformation(
                        **dataclasses.asdict(coordinate_transformation)
                    ),
                    **mapping_properties_kwargs,
                )
            )

    def export_shell_composite_definitions(self, path: _PATH) -> None:
        """
        Export the lay-up of the shell as HDF5 used by DPF Composites or Mechanical.

        Parameters
        ----------
        path:
            File path. Eg. /tmp/ACPCompositeDefinitions.h5
        """
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                self._get_stub().SaveShellCompositeDefinitions(
                    model_pb2.SaveShellCompositeDefinitionsRequest(
                        resource_path=self._resource_path, path=export_path
                    )
                )

    @supported_since("25.1")
    def import_materials(
        self,
        matml_path: _PATH,
        *,
        material_apdl_path: _PATH | None = None,
    ) -> None:
        """
        Import materials from a MatML file.

        Import materials from a ``MatML.xml`` (Engineering Data) file.

        Optionally, a material APDL file can be defined. This is a pre-generated
        solver snippet, needed in case of variable materials or non-standard
        material models. The snippet is used when exporting solid models or
        surface section cuts in the CDB format.

        Parameters
        ----------
        matml_path:
            File path to the MatML file.
        material_apdl_path:
            File path to the material APDL file.
        """
        material_stub = material_pb2_grpc.ObjectServiceStub(self._channel)
        collection_path = CollectionPath(
            value=rp_join(self._resource_path.value, Material._COLLECTION_LABEL)
        )

        with wrap_grpc_errors():
            material_stub.ImportMaterialFiles(
                material_pb2.ImportMaterialFilesRequest(
                    collection_path=collection_path,
                    matml_path=self._server_wrapper.auto_upload(matml_path),
                    material_apdl_path=self._server_wrapper.auto_upload(
                        material_apdl_path, allow_none=True
                    ),
                )
            )

    def export_materials(self, path: _PATH) -> None:
        """
        Write materials to a XML (MatML) file.

        The XML file is required for the post-processing with DPF or can be loaded by
        Engineering Data under WB.

        Parameters
        ----------
        path:
            File path. E.g. /tmp/acp_materials.xml
        """
        material_stub = material_pb2_grpc.ObjectServiceStub(self._channel)
        collection_path = CollectionPath(
            value=rp_join(self._resource_path.value, Material._COLLECTION_LABEL)
        )
        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                material_stub.SaveToFile(
                    material_pb2.SaveToFileRequest(
                        collection_path=collection_path,
                        path=export_path,
                        format=material_pb2.SaveToFileRequest.ANSYS_XML,
                    )
                )

    @supported_since("25.1")
    def export_modeling_ply_geometries(
        self,
        path: _PATH,
        *,
        modeling_plies: Iterable[ModelingPly] | None = None,
        format: PlyGeometryExportFormat = PlyGeometryExportFormat.STEP,
        offset_type: OffsetType = OffsetType.MIDDLE_OFFSET,
        include_surface: bool = True,
        include_boundary: bool = True,
        include_first_material_direction: bool = True,
        include_second_material_direction: bool = True,
        arrow_length: float | None = None,
        arrow_type: ArrowType = ArrowType.NO_ARROW,
    ) -> None:
        """
        Write ply geometries to a STEP, IGES, or STL file.

        Parameters
        ----------
        path :
            File path to save the geometries to.
        modeling_plies :
            List of modeling plies whose geometries should be exported. If not
            provided, the geometries of all modeling plies in the model are exported.
        format :
            Format of the created file. Can be one of ``"STEP"``, ``"IGES"``,
            or ``"STL"``.
        offset_type :
            Determines how the ply offset is calculated. Can be one of
            ``"NO_OFFSET"``, ``"BOTTOM_OFFSET"``, ``"MIDDLE_OFFSET"``, or
            ``"TOP_OFFSET"``.
        include_surface :
            Whether to include the ply surface in the exported geometry.
        include_boundary :
            Whether to include the ply boundary in the exported geometry.
        include_first_material_direction :
            Whether to include the first material direction in the exported geometry.
        include_second_material_direction :
            Whether to include the second material direction in the exported geometry.
        arrow_length :
            Size of the arrow used to represent the material directions. By default, the
            square root of the average element area is used.
        arrow_type :
            Type of the arrow used to represent the material directions. Can be
            one of ``"NO_ARROW"``, ``"HALF_ARROW"``, or ``"STANDARD_ARROW"``.
        """
        if modeling_plies is None:
            modeling_plies = [
                ply
                for modeling_group in self.modeling_groups.values()
                for ply in modeling_group.modeling_plies.values()
            ]
        mp_resource_paths = [ply._resource_path for ply in modeling_plies]

        modeling_ply_stub = modeling_ply_pb2_grpc.ObjectServiceStub(self._channel)

        if arrow_length is None:
            arrow_length = np.sqrt(self.average_element_size)

        with self._server_wrapper.auto_download(path) as export_path:
            with wrap_grpc_errors():
                modeling_ply_stub.ExportGeometries(
                    ply_geometry_export_pb2.ExportGeometriesRequest(
                        path=export_path,
                        plies=mp_resource_paths,
                        options=ply_geometry_export_pb2.ExportOptions(
                            format=typing.cast(
                                typing.Any, ply_geometry_export_format_to_pb(format)
                            ),
                            offset_type=typing.cast(typing.Any, offset_type_to_pb(offset_type)),
                            include_surface=include_surface,
                            include_boundary=include_boundary,
                            include_first_material_direction=include_first_material_direction,
                            include_second_material_direction=include_second_material_direction,
                            arrow_length=arrow_length,
                            arrow_type=typing.cast(typing.Any, arrow_type_to_pb(arrow_type)),
                        ),
                    )
                )

    create_material = define_create_method(
        Material, func_name="create_material", parent_class_name="Model", module_name=__module__
    )
    materials = define_mutable_mapping(Material, material_pb2_grpc.ObjectServiceStub)

    create_fabric = define_create_method(
        Fabric, func_name="create_fabric", parent_class_name="Model", module_name=__module__
    )
    fabrics = define_mutable_mapping(Fabric, fabric_pb2_grpc.ObjectServiceStub)

    create_stackup = define_create_method(
        Stackup, func_name="create_stackup", parent_class_name="Model", module_name=__module__
    )
    stackups = define_mutable_mapping(Stackup, stackup_pb2_grpc.ObjectServiceStub)

    create_sublaminate = define_create_method(
        SubLaminate,
        func_name="create_sublaminate",
        parent_class_name="Model",
        module_name=__module__,
    )
    sublaminates = define_mutable_mapping(SubLaminate, sublaminate_pb2_grpc.ObjectServiceStub)

    create_element_set = define_create_method(
        ElementSet,
        func_name="create_element_set",
        parent_class_name="Model",
        module_name=__module__,
    )
    element_sets = define_mutable_mapping(ElementSet, element_set_pb2_grpc.ObjectServiceStub)

    create_edge_set = define_create_method(
        EdgeSet, func_name="create_edge_set", parent_class_name="Model", module_name=__module__
    )
    edge_sets = define_mutable_mapping(EdgeSet, edge_set_pb2_grpc.ObjectServiceStub)

    create_cad_geometry = define_create_method(
        CADGeometry,
        func_name="create_cad_geometry",
        parent_class_name="Model",
        module_name=__module__,
    )
    cad_geometries = define_mutable_mapping(CADGeometry, cad_geometry_pb2_grpc.ObjectServiceStub)

    create_virtual_geometry = define_create_method(
        VirtualGeometry,
        func_name="create_virtual_geometry",
        parent_class_name="Model",
        module_name=__module__,
    )
    virtual_geometries = define_mutable_mapping(
        VirtualGeometry, virtual_geometry_pb2_grpc.ObjectServiceStub
    )

    create_rosette = define_create_method(
        Rosette, func_name="create_rosette", parent_class_name="Model", module_name=__module__
    )
    rosettes = define_mutable_mapping(Rosette, rosette_pb2_grpc.ObjectServiceStub)

    create_lookup_table_1d = define_create_method(
        LookUpTable1D,
        func_name="create_lookup_table_1d",
        parent_class_name="Model",
        module_name=__module__,
    )
    lookup_tables_1d = define_mutable_mapping(
        LookUpTable1D, lookup_table_1d_pb2_grpc.ObjectServiceStub
    )

    create_lookup_table_3d = define_create_method(
        LookUpTable3D,
        func_name="create_lookup_table_3d",
        parent_class_name="Model",
        module_name=__module__,
    )
    lookup_tables_3d = define_mutable_mapping(
        LookUpTable3D, lookup_table_3d_pb2_grpc.ObjectServiceStub
    )

    create_parallel_selection_rule = define_create_method(
        ParallelSelectionRule,
        func_name="create_parallel_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    parallel_selection_rules = define_mutable_mapping(
        ParallelSelectionRule, parallel_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_cylindrical_selection_rule = define_create_method(
        CylindricalSelectionRule,
        func_name="create_cylindrical_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    cylindrical_selection_rules = define_mutable_mapping(
        CylindricalSelectionRule, cylindrical_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_spherical_selection_rule = define_create_method(
        SphericalSelectionRule,
        func_name="create_spherical_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    spherical_selection_rules = define_mutable_mapping(
        SphericalSelectionRule, spherical_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_tube_selection_rule = define_create_method(
        TubeSelectionRule,
        func_name="create_tube_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    tube_selection_rules = define_mutable_mapping(
        TubeSelectionRule, tube_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_cut_off_selection_rule = define_create_method(
        CutOffSelectionRule,
        func_name="create_cut_off_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    cut_off_selection_rules = define_mutable_mapping(
        CutOffSelectionRule, cutoff_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_geometrical_selection_rule = define_create_method(
        GeometricalSelectionRule,
        func_name="create_geometrical_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    geometrical_selection_rules = define_mutable_mapping(
        GeometricalSelectionRule, geometrical_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_variable_offset_selection_rule = define_create_method(
        VariableOffsetSelectionRule,
        func_name="create_variable_offset_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    variable_offset_selection_rules = define_mutable_mapping(
        VariableOffsetSelectionRule, variable_offset_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_boolean_selection_rule = define_create_method(
        BooleanSelectionRule,
        func_name="create_boolean_selection_rule",
        parent_class_name="Model",
        module_name=__module__,
    )
    boolean_selection_rules = define_mutable_mapping(
        BooleanSelectionRule, boolean_selection_rule_pb2_grpc.ObjectServiceStub
    )

    create_oriented_selection_set = define_create_method(
        OrientedSelectionSet,
        func_name="create_oriented_selection_set",
        parent_class_name="Model",
        module_name=__module__,
    )
    oriented_selection_sets = define_mutable_mapping(
        OrientedSelectionSet, oriented_selection_set_pb2_grpc.ObjectServiceStub
    )

    create_modeling_group = define_create_method(
        ModelingGroup,
        func_name="create_modeling_group",
        parent_class_name="Model",
        module_name=__module__,
    )
    modeling_groups = define_mutable_mapping(
        ModelingGroup, modeling_group_pb2_grpc.ObjectServiceStub
    )

    create_imported_modeling_group = define_create_method(
        ImportedModelingGroup,
        func_name="create_imported_modeling_group",
        parent_class_name="Model",
        module_name=__module__,
    )
    imported_modeling_groups = define_mutable_mapping(
        ImportedModelingGroup, imported_modeling_group_pb2_grpc.ObjectServiceStub
    )

    create_sampling_point = define_create_method(
        SamplingPoint,
        func_name="create_sampling_point",
        parent_class_name="Model",
        module_name=__module__,
    )
    sampling_points = define_mutable_mapping(
        SamplingPoint, sampling_point_pb2_grpc.ObjectServiceStub
    )

    create_section_cut = define_create_method(
        SectionCut,
        func_name="create_section_cut",
        parent_class_name="Model",
        module_name=__module__,
    )
    section_cuts = define_mutable_mapping(SectionCut, section_cut_pb2_grpc.ObjectServiceStub)

    create_solid_model = define_create_method(
        SolidModel,
        func_name="create_solid_model",
        parent_class_name="Model",
        module_name=__module__,
    )
    solid_models = define_mutable_mapping(SolidModel, solid_model_pb2_grpc.ObjectServiceStub)

    create_imported_solid_model = define_create_method(
        ImportedSolidModel,
        func_name="create_imported_solid_model",
        parent_class_name="Model",
        module_name=__module__,
    )
    imported_solid_models = define_mutable_mapping(
        ImportedSolidModel, imported_solid_model_pb2_grpc.ObjectServiceStub
    )

    create_sensor = define_create_method(
        Sensor, func_name="create_sensor", parent_class_name="Model", module_name=__module__
    )
    sensors = define_mutable_mapping(Sensor, sensor_pb2_grpc.ObjectServiceStub)

    create_field_definition = define_create_method(
        FieldDefinition,
        func_name="create_field_definition",
        parent_class_name="Model",
        module_name=__module__,
    )
    field_definitions = define_mutable_mapping(
        FieldDefinition, field_definition_pb2_grpc.ObjectServiceStub
    )

    mesh = full_mesh_property
    shell_mesh = shell_mesh_property
    solid_mesh = solid_mesh_property
    elemental_data = elemental_data_property(ModelElementalData)
    nodal_data = nodal_data_property(ModelNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.api.acp.v0 import (
    cut_off_geometry_pb2,
    cut_off_material_pb2,
    cutoff_selection_rule_pb2,
    drop_off_material_pb2,
    edge_set_pb2,
    enum_types_pb2,
    extrusion_guide_pb2,
    geometrical_selection_rule_pb2,
    imported_modeling_ply_pb2,
    layup_mapping_object_pb2,
    lookup_table_3d_pb2,
    lookup_table_column_type_pb2,
    mesh_query_pb2,
    modeling_ply_pb2,
    ply_material_pb2,
    rosette_pb2,
    section_cut_pb2,
    sensor_pb2,
    snap_to_geometry_pb2,
    solid_model_pb2,
    unit_system_pb2,
    virtual_geometry_pb2,
)

from ._grpc_helpers.enum_wrapper import wrap_to_string_enum

__all__ = [
    "ArrowType",
    "BaseElementMaterialHandling",
    "BooleanOperationType",
    "CutOffGeometryOrientationType",
    "CutOffMaterialHandling",
    "CutOffRuleType",
    "DrapingMaterialModel",
    "DrapingType",
    "DropOffMaterialHandling",
    "DropOffType",
    "EdgeSetType",
    "ElementalDataType",
    "ElementTechnology",
    "ExtrusionGuideType",
    "ExtrusionMethod",
    "ExtrusionType",
    "GeometricalRuleType",
    "ImportedPlyDrapingType",
    "ImportedPlyOffsetType",
    "ImportedPlyThicknessType",
    "IntersectionType",
    "LookUpTable3DInterpolationAlgorithm",
    "LookUpTableColumnValueType",
    "NodalDataType",
    "OffsetType",
    "PhysicalDimension",
    "PlyCutOffType",
    "PlyGeometryExportFormat",
    "PlyType",
    "ReinforcingBehavior",
    "RosetteSelectionMethod",
    "RosetteType",
    "SectionCutCDBExportType",
    "SectionCutType",
    "SensorType",
    "SnapToGeometryOrientationType",
    "SolidModelExportFormat",
    "SolidModelOffsetDirectionType",
    "SolidModelSkinExportFormat",
    "Status",
    "StressStateType",
    "SymmetryType",
    "ThicknessFieldType",
    "ThicknessType",
    "UnitSystemType",
    "VirtualGeometryDimension",
]

(Status, status_type_to_pb, status_type_from_pb) = wrap_to_string_enum(
    "Status",
    enum_types_pb2.StatusType,
    module=__name__,
    value_converter=lambda val: val,
    doc="Options for the update status of an object.",
)
(
    RosetteSelectionMethod,
    rosette_selection_method_to_pb,
    rosette_selection_method_from_pb,
) = wrap_to_string_enum(
    "RosetteSelectionMethod",
    enum_types_pb2.RosetteSelectionMethod,
    module=__name__,
    doc="Options for how the rosette is selected in oriented selection sets.",
)

(
    CutOffMaterialHandling,
    cut_off_material_type_to_pb,
    cut_off_material_type_from_pb,
) = wrap_to_string_enum(
    "CutOffMaterialHandling",
    cut_off_material_pb2.MaterialHandlingType,
    module=__name__,
    doc="Options for how cut-off material is selected.",
)

(
    DropOffMaterialHandling,
    drop_off_material_type_to_pb,
    drop_off_material_type_from_pb,
) = wrap_to_string_enum(
    "DropOffMaterialHandling",
    drop_off_material_pb2.MaterialHandlingType,
    module=__name__,
    doc="Options for how drop-off material is selected.",
)

(
    DrapingType,
    draping_type_to_pb,
    draping_type_from_pb,
) = wrap_to_string_enum(
    "DrapingType",
    ply_material_pb2.DrapingType,
    module=__name__,
    doc="Options for the draping algorithm used.",
)

(
    ImportedPlyDrapingType,
    imported_ply_draping_type_to_pb,
    imported_ply_draping_type_from_pb,
) = wrap_to_string_enum(
    "ImportedPlyDrapingType",
    ply_material_pb2.DrapingType,
    module=__name__,
    doc="Options for the draping algorithm used.",
    explicit_value_list=(
        ply_material_pb2.DrapingType.NO_DRAPING,
        ply_material_pb2.DrapingType.TABULAR_VALUES,
    ),
)

(
    DrapingMaterialModel,
    draping_material_type_to_pb,
    draping_material_type_from_pb,
) = wrap_to_string_enum(
    "DrapingMaterialModel",
    ply_material_pb2.DrapingMaterialType,
    module=__name__,
    doc="Options for the material type used in the draping algorithm.",
)

(
    SymmetryType,
    symmetry_type_to_pb,
    symmetry_type_from_pb,
) = wrap_to_string_enum(
    "SymmetryType",
    ply_material_pb2.SymmetryType,
    module=__name__,
    doc="Options for the symmetry of stackups or sublaminates.",
)

(
    EdgeSetType,
    edge_set_type_to_pb,
    edge_set_type_from_pb,
) = wrap_to_string_enum(
    "EdgeSetType",
    edge_set_pb2.EdgeSetType,
    module=__name__,
    doc="Options for how an edge set is defined.",
)

(
    RosetteType,
    rosette_type_to_pb,
    rosette_type_from_pb,
) = wrap_to_string_enum(
    "RosetteType",
    rosette_pb2.Type,
    module=__name__,
    doc="Options for the type of a rosette.",
)

(
    PlyType,
    ply_type_to_pb,
    ply_type_from_pb,
) = wrap_to_string_enum(
    "PlyType",
    enum_types_pb2.PlyType,
    module=__name__,
    doc="Options for the material type of a ply.",
)

(
    BooleanOperationType,
    boolean_operation_type_to_pb,
    boolean_operation_type_from_pb,
) = wrap_to_string_enum(
    "BooleanOperationType",
    enum_types_pb2.BooleanOperationType,
    module=__name__,
    doc="Options for combining selection rules.",
)

(OffsetType, offset_type_to_pb, _) = wrap_to_string_enum(
    "OffsetType",
    enum_types_pb2.OffsetType,
    module=__name__,
    doc="Options for the ply offset type.",
)

(ArrowType, arrow_type_to_pb, _) = wrap_to_string_enum(
    "ArrowType",
    enum_types_pb2.ArrowType,
    module=__name__,
    doc="Options for the type of arrow to be created for directions in the ply geometry export.",
)

(
    UnitSystemType,
    unit_system_type_to_pb,
    unit_system_type_from_pb,
) = wrap_to_string_enum(
    "UnitSystemType",
    unit_system_pb2.UnitSystemType,
    module=__name__,
    doc="Available choices for the unit system.",
    # When loading from a file, the value 'from_file' is more descriptive than 'undefined',
    # so we add an alias for it.
    extra_aliases={"undefined": ("FROM_FILE", "from_file")},
)

(
    PhysicalDimension,
    physical_dimension_to_pb,
    physical_dimension_from_pb,
) = wrap_to_string_enum(
    "PhysicalDimension",
    unit_system_pb2.DimensionType,
    module=__name__,
    doc="Options for the dimension (time, length, currency, ...) of data.",
)

(
    ElementalDataType,
    elemental_data_type_to_pb,
    elemental_data_type_from_pb,
) = wrap_to_string_enum(
    "ElementalDataType",
    mesh_query_pb2.ElementalDataType,
    module=__name__,
    # The enum value names in the '.proto' definition are prefixed with 'ELEMENT_',
    # since they must be unique within the 'mesh_query.proto' file and would
    # otherwise conflict with the 'NodalDataType' enum.
    # For the Python enum, we remove the prefix and convert the values to
    # lowercase.
    key_converter=lambda val: val.split("_", 1)[1],
    value_converter=lambda val: val.split("_", 1)[1].lower(),
    doc="Options for the type of per-element data.",
)
(
    NodalDataType,
    nodal_data_type_to_pb,
    nodal_data_type_from_pb,
) = wrap_to_string_enum(
    "NodalDataType",
    mesh_query_pb2.NodalDataType,
    module=__name__,
    # The enum value names in the '.proto' definition are prefixed with 'NODE_',
    # since they must be unique within the 'mesh_query.proto' file and would
    # otherwise conflict with the 'ElementalDataType' enum.
    # For the Python enum, we remove the prefix and convert the values to
    # lowercase.
    key_converter=lambda val: val.split("_", 1)[1],
    value_converter=lambda val: val.split("_", 1)[1].lower(),
    doc="Options for the type of per-node data.",
)

(
    LookUpTableColumnValueType,
    lookup_table_column_value_type_to_pb,
    lookup_table_column_value_type_from_pb,
) = wrap_to_string_enum(
    "LookUpTableColumnValueType",
    lookup_table_column_type_pb2.ValueType,
    module=__name__,
    doc=(
        "Options for the column type (data location, scalar data, directional data) "
        "of look-up table columns."
    ),
)
(
    LookUpTable3DInterpolationAlgorithm,
    lookup_table_3d_interpolation_algorithm_to_pb,
    lookup_table_3d_interpolation_algorithm_from_pb,
) = wrap_to_string_enum(
    "LookUpTable3DInterpolationAlgorithm",
    lookup_table_3d_pb2.InterpolationAlgorithm,
    module=__name__,
    doc="Options for the interpolation algorithm used for 3D look-up tables.",
)
(
    SensorType,
    sensor_type_to_pb,
    sensor_type_from_pb,
) = wrap_to_string_enum(
    "SensorType", sensor_pb2.SensorType, module=__name__, doc="Options for the type of sensor."
)

(
    VirtualGeometryDimension,
    virtual_geometry_dimension_to_pb,
    virtual_geometry_dimension_from_pb,
) = wrap_to_string_enum(
    "VirtualGeometryDimension",
    virtual_geometry_pb2.Dimension,
    module=__name__,
    doc="Options for the dimension of a virtual geometry.",
)

(
    CutOffRuleType,
    cut_off_rule_type_to_pb,
    cut_off_rule_type_from_pb,
) = wrap_to_string_enum(
    "CutOffRuleType",
    cutoff_selection_rule_pb2.CutoffRuleType,
    module=__name__,
    doc="Options for how a cut off rule is defined.",
)

(
    PlyCutOffType,
    ply_cut_off_type_to_pb,
    ply_cut_off_type_from_pb,
) = wrap_to_string_enum(
    "PlyCutOffType",
    cutoff_selection_rule_pb2.PlyCutoffType,
    module=__name__,
    doc="Options for how ply cut-off is computed.",
)

(
    GeometricalRuleType,
    geometrical_rule_type_to_pb,
    geometrical_rule_type_from_pb,
) = wrap_to_string_enum(
    "GeometricalRuleType",
    geometrical_selection_rule_pb2.GeometricalRuleType,
    module=__name__,
    doc="Options for how a geometrical selection rule is defined.",
)
(
    ThicknessType,
    thickness_type_to_pb,
    thickness_type_from_pb,
) = wrap_to_string_enum(
    "ThicknessType",
    modeling_ply_pb2.ThicknessType,
    module=__name__,
    doc="Options for how ply thickness is defined.",
)

(
    ImportedPlyThicknessType,
    imported_ply_thickness_type_to_pb,
    imported_ply_thickness_type_from_pb,
) = wrap_to_string_enum(
    "ImportedPlyThicknessType",
    modeling_ply_pb2.ThicknessType,
    module=__name__,
    doc="Options for how ply thickness is defined.",
    explicit_value_list=(
        modeling_ply_pb2.ThicknessType.NOMINAL,
        modeling_ply_pb2.ThicknessType.FROM_TABLE,
    ),
)

(
    ThicknessFieldType,
    thickness_field_type_to_pb,
    thickness_field_type_from_pb,
) = wrap_to_string_enum(
    "ThicknessFieldType",
    modeling_ply_pb2.ThicknessFieldType,
    module=__name__,
    doc="Options for how thickness from a table is defined.",
)

(PlyGeometryExportFormat, ply_geometry_export_format_to_pb, _) = wrap_to_string_enum(
    "PlyGeometryExportFormat",
    enum_types_pb2.FileFormat,
    module=__name__,
    doc="Options for the file format of the ply geometry export.",
    explicit_value_list=(
        enum_types_pb2.FileFormat.STEP,
        enum_types_pb2.FileFormat.IGES,
        enum_types_pb2.FileFormat.STL,
    ),
)

(
    ImportedPlyOffsetType,
    imported_ply_offset_type_to_pb,
    imported_ply_offset_type_from_pb,
) = wrap_to_string_enum(
    "ImportedPlyOffsetType",
    enum_types_pb2.OffsetType,
    module=__name__,
    doc="Options for the definition of the offset.",
    explicit_value_list=(
        enum_types_pb2.OffsetType.MIDDLE_OFFSET,
        enum_types_pb2.OffsetType.BOTTOM_OFFSET,
        enum_types_pb2.OffsetType.TOP_OFFSET,
    ),
)

(
    MeshImportType,
    mesh_import_type_to_pb,
    mesh_import_type_from_pb,
) = wrap_to_string_enum(
    "MeshImportType",
    imported_modeling_ply_pb2.MeshImportType,
    module=__name__,
    doc="Options for the definition of the source of the imported mesh.",
)

(ExtrusionType, extrusion_type_to_pb, extrusion_type_from_pb) = wrap_to_string_enum(
    "ExtrusionType",
    section_cut_pb2.ExtrusionType,
    module=__name__,
    doc="Extrusion method used in a section cut.",
)

(SectionCutType, section_cut_type_to_pb, section_cut_type_from_pb) = wrap_to_string_enum(
    "SectionCutType",
    section_cut_pb2.SectionCutType,
    module=__name__,
    doc="Determines whether the section cut is extruded by modeling ply, production ply, or analysis ply.",
)

(IntersectionType, intersection_type_to_pb, intersection_type_from_pb) = wrap_to_string_enum(
    "IntersectionType",
    section_cut_pb2.IntersectionType,
    module=__name__,
    doc="Determines how the intersection is computed for wireframe section cuts.",
)

SectionCutCDBExportType, section_cut_cdb_export_type_to_pb, _ = wrap_to_string_enum(
    "SectionCutCDBExportType",
    section_cut_pb2.CDBExportType,
    module=__name__,
    doc="Determines the type of section cut model to be exported to ANSYS Mechanical APDL.",
)

(ExtrusionMethod, extrusion_method_type_to_pb, extrusion_method_type_from_pb) = wrap_to_string_enum(
    "ExtrusionMethod",
    solid_model_pb2.ExtrusionMethodType,
    module=__name__,
    doc="Extrusion method used in a solid model.",
)

(ExtrusionGuideType, extrusion_guide_type_to_pb, extrusion_guide_type_from_pb) = (
    wrap_to_string_enum(
        "ExtrusionGuideType",
        extrusion_guide_pb2.ExtrusionGuideType,
        module=__name__,
        doc="Extrusion guide type used in an extrusion guide (solid model).",
    )
)

(SolidModelOffsetDirectionType, offset_direction_type_to_pb, offset_direction_type_from_pb) = (
    wrap_to_string_enum(
        "SolidModelOffsetDirectionType",
        solid_model_pb2.OffsetDirectionType,
        module=__name__,
        doc=(
            "Determines how the offset direction is evaluated in a solid model. With "
            "``SURFACE_NORMAL``, the offset direction is re-evaluated based on the "
            "surface of the solid. With ``SHELL_NORMAL``, the direction is based on the "
            "shell surface."
        ),
    )
)
(DropOffType, drop_off_type_to_pb, drop_off_type_from_pb) = wrap_to_string_enum(
    "DropOffType",
    solid_model_pb2.DropOffType,
    module=__name__,
    doc="Determines whether the drop off in solid models is inside or outside the ply boundary.",
)

SolidModelExportFormat, solid_model_export_format_to_pb, _ = wrap_to_string_enum(
    "SolidModelExportFormat",
    enum_types_pb2.FileFormat,
    module=__name__,
    value_converter=lambda val: val.lower().replace("_", ":"),
    doc="Options for the export format of solid models.",
    explicit_value_list=(
        enum_types_pb2.FileFormat.ANSYS_H5,
        enum_types_pb2.FileFormat.ANSYS_CDB,
    ),
)

SolidModelSkinExportFormat, solid_model_skin_export_format_to_pb, _ = wrap_to_string_enum(
    "SolidModelSkinExportFormat",
    enum_types_pb2.FileFormat,
    module=__name__,
    value_converter=lambda val: val.lower().replace("_", ":"),
    doc="Options for the export format of solid model skins.",
    explicit_value_list=(
        enum_types_pb2.FileFormat.ANSYS_CDB,
        enum_types_pb2.FileFormat.STEP,
        enum_types_pb2.FileFormat.IGES,
        enum_types_pb2.FileFormat.STL,
    ),
)


def _prefix_undefined(value: str) -> str:
    if value == "UNDEFINED":
        return "_UNDEFINED"
    return value


(
    SnapToGeometryOrientationType,
    snap_to_geometry_orientation_type_to_pb,
    snap_to_geometry_orientation_type_from_pb,
) = wrap_to_string_enum(
    "SnapToGeometryOrientationType",
    snap_to_geometry_pb2.OrientationType,
    module=__name__,
    key_converter=_prefix_undefined,
    value_converter=lambda val: _prefix_undefined(val).lower(),
    doc=(
        "Determines which layup face a snap-to geometry is applies to. Note that the "
        "``_UNDEFINED`` option should not be used. It is equivalent to using "
        "``BOTTOM``, and included only for compatibility with existing models."
    ),
)
(
    CutOffGeometryOrientationType,
    cut_off_geometry_orientation_type_to_pb,
    cut_off_geometry_orientation_type_from_pb,
) = wrap_to_string_enum(
    "CutOffGeometryOrientationType",
    cut_off_geometry_pb2.OrientationType,
    module=__name__,
    doc="Determines the orientation of a cut-off geometry.",
)

ElementTechnology, element_technology_to_pb, element_technology_from_pb = wrap_to_string_enum(
    "ElementTechnology",
    layup_mapping_object_pb2.ElementTechnology,
    module=__name__,
    doc=("Options for the element technology used in a layup mapping object."),
)

ReinforcingBehavior, reinforcing_behavior_to_pb, reinforcing_behavior_from_pb = wrap_to_string_enum(
    "ReinforcingBehavior",
    layup_mapping_object_pb2.ReinforcingBehavior,
    module=__name__,
    doc=(
        "Specifies whether the reinforcing elements carry tension and compression load, or only one of them."
    ),
)

(
    BaseElementMaterialHandling,
    base_element_material_handling_to_pb,
    base_element_material_handling_from_pb,
) = wrap_to_string_enum(
    "BaseElementMaterialHandling",
    layup_mapping_object_pb2.BaseElementMaterialHandlingType,
    module=__name__,
    doc=(
        "Determines how the base material is handled where it intersects with a reinforcing element."
    ),
)
StressStateType, stress_state_type_to_pb, stress_state_type_from_pb = wrap_to_string_enum(
    "StressStateType",
    layup_mapping_object_pb2.StressStateType,
    module=__name__,
    doc="Specifies if the reinforcing elements should behave like a link, membrane, or shell "
    "element (with or without bending).",
)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses
from typing import Any

from ansys.api.acp.v0 import (
    analysis_ply_pb2_grpc,
    cut_off_geometry_pb2_grpc,
    extrusion_guide_pb2_grpc,
    interface_layer_pb2_grpc,
    snap_to_geometry_pb2_grpc,
    solid_element_set_pb2_grpc,
    solid_model_pb2,
    solid_model_pb2_grpc,
)

from .._utils.property_protocols import ReadOnlyProperty, ReadWriteProperty
from ._elemental_or_nodal_data import (
    ElementalData,
    NodalData,
    VectorData,
    elemental_data_property,
    nodal_data_property,
)
from ._grpc_helpers.linked_object_list import (
    define_linked_object_list,
    define_polymorphic_linked_object_list,
)
from ._grpc_helpers.mapping import (
    define_create_method,
    define_mutable_mapping,
    get_read_only_collection_property,
)
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from ._solid_model_export import SolidModelExportMixin
from .analysis_ply import AnalysisPly
from .base import (
    CreatableTreeObject,
    IdTreeObject,
    TreeObjectAttributeWithCache,
    nested_grpc_object_property,
)
from .cut_off_geometry import CutOffGeometry
from .edge_set import EdgeSet
from .element_set import ElementSet
from .enums import (
    DropOffType,
    ExtrusionMethod,
    SolidModelOffsetDirectionType,
    drop_off_type_from_pb,
    drop_off_type_to_pb,
    extrusion_method_type_from_pb,
    extrusion_method_type_to_pb,
    offset_direction_type_from_pb,
    offset_direction_type_to_pb,
    status_type_from_pb,
)
from .extrusion_guide import ExtrusionGuide
from .interface_layer import InterfaceLayer
from .material import Material
from .modeling_ply import ModelingPly
from .object_registry import register
from .oriented_selection_set import OrientedSelectionSet
from .snap_to_geometry import SnapToGeometry
from .solid_element_set import SolidElementSet

__all__ = [
    "SolidModel",
    "DropOffSettings",
    "SolidModelExportSettings",
    "SolidModelElementalData",
    "SolidModelNodalData",
]


@dataclasses.dataclass
class SolidModelElementalData(ElementalData):
    """Represents elemental data for a Solid Model."""

    normal: VectorData | None = None


@dataclasses.dataclass
class SolidModelNodalData(NodalData):
    """Represents nodal data for a Solid Model."""


@mark_grpc_properties
class DropOffSettings(TreeObjectAttributeWithCache):
    """Defines the drop-off settings for a solid model.

    Parameters
    ----------
    drop_off_type :
        Determines whether the ply's drop-off is inside or outside the boundary
        of the ply.
    disable_drop_offs_on_bottom :
        Whether to remove drop-offs on the bottom surface of the laminate.
    drop_off_disabled_on_bottom_sets :
        Element sets or oriented selection sets on which drop-offs at the bottom
        surface are disabled.
    disable_drop_offs_on_top :
        Whether to remove drop-offs on the top surface of the laminate.
    drop_off_disabled_on_top_sets :
        Element sets or oriented selection sets on which drop-offs at the top
        surface are disabled.
    connect_butt_joined_plies :
        Prevent an element drop-off of two adjacent, sequential plies in the same
        modeling group.
    """

    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        drop_off_type: DropOffType = DropOffType.INSIDE_PLY,
        disable_drop_offs_on_bottom: bool = False,
        drop_off_disabled_on_bottom_sets: Iterable[ElementSet | OrientedSelectionSet] = (),
        disable_drop_offs_on_top: bool = False,
        drop_off_disabled_on_top_sets: Iterable[ElementSet | OrientedSelectionSet] = (),
        connect_butt_joined_plies: bool = True,
        _parent_object: SolidModel | None = None,
        _pb_object: Any | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(
            _parent_object=_parent_object,
            _pb_object=_pb_object,
            _attribute_path=_attribute_path,
        )
        # The '__init__' method can be called either with the explicit values
        # defined below, or from a parent object or protobuf object. In the case
        # where a parent object or protobuf object is provided, the explicit values
        # must not be set, otherwise the default values will override the existing
        # values.
        if _parent_object is None and _pb_object is None:
            self.drop_off_type = drop_off_type
            self.disable_drop_offs_on_bottom = disable_drop_offs_on_bottom
            self.drop_off_disabled_on_bottom_sets = drop_off_disabled_on_bottom_sets
            self.disable_drop_offs_on_top = disable_drop_offs_on_top
            self.drop_off_disabled_on_top_sets = drop_off_disabled_on_top_sets
            self.connect_butt_joined_plies = connect_butt_joined_plies

    @classmethod
    def _create_default_pb_object(self) -> solid_model_pb2.DropOffSettings:
        # There's no need to define the 'real' default values here, since
        # this method is only called when the object is created from scratch.
        # In that case, the '__init__' method will be called with the default
        # values defined there.
        # NOTE dgresch Oct'24: I'm not sure if '_create_default_pb_object' is
        # the right abstraction overall.
        # The concept _could_ be useful for avoiding the duplicate logic of
        # checking '_parent_object is None and _pb_object is None' (this class's
        # '__init__', and the 'TreeObjectAttribute.__init__').
        # But, it would need to be extended to allow also _overriding_ the defaults
        # on calling __init__.
        return solid_model_pb2.DropOffSettings()

    drop_off_type = grpc_data_property(
        "drop_off_type",
        from_protobuf=drop_off_type_from_pb,
        to_protobuf=drop_off_type_to_pb,
    )

    disable_drop_offs_on_bottom: ReadWriteProperty[bool, bool] = grpc_data_property(
        "disable_dropoffs_on_bottom"
    )
    drop_off_disabled_on_bottom_sets = define_polymorphic_linked_object_list(
        "dropoff_disabled_on_bottom_sets", allowed_types=(ElementSet, OrientedSelectionSet)
    )

    disable_drop_offs_on_top: ReadWriteProperty[bool, bool] = grpc_data_property(
        "disable_dropoffs_on_top"
    )
    drop_off_disabled_on_top_sets = define_polymorphic_linked_object_list(
        "dropoff_disabled_on_top_sets", allowed_types=(ElementSet, OrientedSelectionSet)
    )

    connect_butt_joined_plies: ReadWriteProperty[bool, bool] = grpc_data_property(
        "connect_butt_joined_plies"
    )


@mark_grpc_properties
class SolidModelExportSettings(TreeObjectAttributeWithCache):
    """Defines the settings for exporting a solid model.

    Parameters
    ----------
    use_default_section_index :
        Use the default start index for sections.
    section_index :
        Custom start index for sections.
        Only used if ``use_default_section_index`` is False.
    use_default_coordinate_system_index :
        Use the default start index for coordinate systems.
    coordinate_system_index :
        Custom start index for coordinate systems.
        Only used if ``use_default_coordinate_system_index`` is False.
    use_default_material_index :
        Use the default start index for materials.
    material_index :
        Custom start index for materials.
        Only used if ``use_default_material_index`` is False.
    use_default_node_index :
        Use the default start index for nodes.
    node_index :
        Custom start index for nodes.
        Only used if ``use_default_node_index`` is False.
    use_default_element_index :
        Use the default start index for elements.
    element_index :
        Custom start index for elements.
        Only used if ``use_default_element_index`` is False.
    use_solsh_elements :
        When True, export linear layered elements as Solsh (Solid190).
    write_degenerated_elements :
        Whether to export drop-off and cut-off elements.
    drop_hanging_nodes :
        When True, the hanging nodes of quadratic solid meshes are dropped.
    use_solid_model_prefix :
        Use the solid model name as a prefix for the exported file.
    transfer_all_sets :
        When True, all element sets and edge sets are exported.
    transferred_element_sets :
        Element sets to be exported.
        Only used if ``transfer_all_sets`` is False.
    transferred_edge_sets :
        Edge sets to be exported.
        Only used if ``transfer_all_sets`` is False.

    """

    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        use_default_section_index: bool = True,
        section_index: int = 0,
        use_default_coordinate_system_index: bool = True,
        coordinate_system_index: int = 0,
        use_default_material_index: bool = True,
        material_index: int = 0,
        use_default_node_index: bool = True,
        node_index: int = 0,
        use_default_element_index: bool = True,
        element_index: int = 0,
        use_solsh_elements: bool = False,
        write_degenerated_elements: bool = True,
        drop_hanging_nodes: bool = True,
        use_solid_model_prefix: bool = True,
        transfer_all_sets: bool = True,
        transferred_element_sets: Iterable[ElementSet] = (),
        transferred_edge_sets: Iterable[EdgeSet] = (),
        _parent_object: SolidModel | None = None,
        _pb_object: Any | None = None,
        _attribute_path: str | None = None,
    ):
        super().__init__(
            _parent_object=_parent_object,
            _pb_object=_pb_object,
            _attribute_path=_attribute_path,
        )
        # See comment on DropOffSettings.__init__ for the logic here.
        if _parent_object is None and _pb_object is None:
            self.use_default_section_index = use_default_section_index
            self.section_index = section_index
            self.use_default_coordinate_system_index = use_default_coordinate_system_index
            self.coordinate_system_index = coordinate_system_index
            self.use_default_material_index = use_default_material_index
            self.material_index = material_index
            self.use_default_node_index = use_default_node_index
            self.node_index = node_index
            self.use_default_element_index = use_default_element_index
            self.element_index = element_index
            self.use_solsh_elements = use_solsh_elements
            self.write_degenerated_elements = write_degenerated_elements
            self.drop_hanging_nodes = drop_hanging_nodes
            self.use_solid_model_prefix = use_solid_model_prefix
            self.transfer_all_sets = transfer_all_sets
            self.transferred_element_sets = transferred_element_sets
            self.transferred_edge_sets = transferred_edge_sets

    @classmethod
    def _create_default_pb_object(self) -> solid_model_pb2.ExportSettings:
        # See comment on DropOffSettings._create_default_pb_object
        return solid_model_pb2.ExportSettings()

    use_default_section_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_section_index"
    )
    section_index: ReadWriteProperty[int, int] = grpc_data_property("section_index")
    use_default_coordinate_system_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_coordinate_system_index"
    )
    coordinate_system_index: ReadWriteProperty[int, int] = grpc_data_property(
        "coordinate_system_index"
    )
    use_default_material_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_material_index"
    )
    material_index: ReadWriteProperty[int, int] = grpc_data_property("material_index")
    use_default_node_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_node_index"
    )
    node_index: ReadWriteProperty[int, int] = grpc_data_property("node_index")
    use_default_element_index: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_default_element_index"
    )
    element_index: ReadWriteProperty[int, int] = grpc_data_property("element_index")
    use_solsh_elements: ReadWriteProperty[bool, bool] = grpc_data_property("use_solsh_elements")
    write_degenerated_elements: ReadWriteProperty[bool, bool] = grpc_data_property(
        "write_degenerated_elements"
    )
    drop_hanging_nodes: ReadWriteProperty[bool, bool] = grpc_data_property("drop_hanging_nodes")
    use_solid_model_prefix: ReadWriteProperty[bool, bool] = grpc_data_property(
        "use_solid_model_prefix"
    )
    transfer_all_sets: ReadWriteProperty[bool, bool] = grpc_data_property("transfer_all_sets")
    transferred_element_sets = define_linked_object_list("transferred_element_sets", ElementSet)
    transferred_edge_sets = define_linked_object_list("transferred_edge_sets", EdgeSet)


@mark_grpc_properties
@register
class SolidModel(SolidModelExportMixin, CreatableTreeObject, IdTreeObject):
    """Instantiate a solid model.

    Parameters
    ----------
    name :
        Name of the solid model.
    active :
        Inactive solid models are not computed, and ignored in the analysis.
    element_sets :
        Element sets or oriented selection sets determining the extent of
        the solid model.
    extrusion_method :
        Determines how plies are bundled in the layered solid elements.
    max_element_thickness :
        Maximum thickness of the layered solid elements. A new element is
        introduced if the thickness exceeds this value.
        Only used if the ``extrusion_method`` is one of ``SPECIFY_THICKNESS``,
        ``MATERIAL_WISE``, or ``SANDWICH_WISE``.
    ply_group_pointers :
        Explicitly defines modeling plies where a new element is introduced.
        Only used if the ``extrusion_method`` is ``USER_DEFINED``.
    offset_direction_type :
        Determines how the extrusion direction is defined. With ``SHELL_NORMAL``,
        the normal direction of the shell is used during the entire extrusion.
        With ``SURFACE_NORMAL``, the offset direction is re-evaluated based
        on the surface of the solid model during the extrusion.
    skip_elements_without_plies :
        If True, elements without plies are automatically removed from the
        region of extrusion. This means that no drop-off elements are created
        for these elements.
    drop_off_material :
        This material is assigned to the layered solid drop-off elements if
        ``drop_off_material_handling`` is set to ``GLOBAL`` in the fabric
        definition.
    cut_off_material :
        This material is assigned to the degenerated solid cut-off elements if
        ``cut_off_material_handling`` is set to ``GLOBAL`` in the fabric
        definition.
    delete_bad_elements :
        If True, a final element check is performed to remove erroneous elements.
    warping_limit :
        Maximum allowable warping limit used in the element shape check. Elements
        with a warping limit exceeding this value are removed.
        Only used if ``delete_bad_elements`` is True.
    minimum_volume :
        Solid elements with a volume smaller or equal to this value are removed.
        With the default value of ``0``, only inverted or zero-volume elements
        are removed.
        Only used if ``delete_bad_elements`` is True.
    drop_off_settings :
        Determines how drop-off elements are handled in the solid model extrusion.
    export_settings :
        Defines the settings for exporting the solid model.

    """

    __slots__: Iterable[str] = tuple()
    _COLLECTION_LABEL = "solid_models"
    _OBJECT_INFO_TYPE = solid_model_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = solid_model_pb2.CreateRequest
    _SUPPORTED_SINCE = "25.1"

    def __init__(
        self,
        *,
        name: str = "SolidModel",
        active: bool = True,
        element_sets: Iterable[ElementSet | OrientedSelectionSet] = (),
        extrusion_method: ExtrusionMethod = ExtrusionMethod.ANALYSIS_PLY_WISE,
        max_element_thickness: float = 1.0,
        ply_group_pointers: Iterable[ModelingPly] = (),
        offset_direction_type: SolidModelOffsetDirectionType = SolidModelOffsetDirectionType.SHELL_NORMAL,
        skip_elements_without_plies: bool = False,
        drop_off_material: Material | None = None,
        cut_off_material: Material | None = None,
        delete_bad_elements: bool = True,
        warping_limit: float = 0.4,
        minimum_volume: float = 0.0,
        drop_off_settings: DropOffSettings = DropOffSettings(),
        export_settings: SolidModelExportSettings = SolidModelExportSettings(),
    ):
        super().__init__(
            name=name,
        )
        self.active = active
        self.element_sets = element_sets
        self.extrusion_method = extrusion_method
        self.max_element_thickness = max_element_thickness
        self.ply_group_pointers = ply_group_pointers
        self.offset_direction_type = offset_direction_type
        self.skip_elements_without_plies = skip_elements_without_plies
        self.drop_off_material = drop_off_material
        self.cut_off_material = cut_off_material
        self.delete_bad_elements = delete_bad_elements
        self.warping_limit = warping_limit
        self.minimum_volume = minimum_volume
        self.drop_off_settings = drop_off_settings
        self.export_settings = export_settings

    def _create_stub(self) -> solid_model_pb2_grpc.ObjectServiceStub:
        return solid_model_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    active: ReadWriteProperty[bool, bool] = grpc_data_property("properties.active")

    element_sets = define_polymorphic_linked_object_list(
        "properties.element_sets", allowed_types=(ElementSet, OrientedSelectionSet)
    )
    extrusion_method = grpc_data_property(
        "properties.extrusion_method",
        from_protobuf=extrusion_method_type_from_pb,
        to_protobuf=extrusion_method_type_to_pb,
    )
    max_element_thickness: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.max_element_thickness"
    )
    ply_group_pointers = define_linked_object_list("properties.ply_group_pointers", ModelingPly)
    offset_direction_type = grpc_data_property(
        "properties.offset_direction",
        from_protobuf=offset_direction_type_from_pb,
        to_protobuf=offset_direction_type_to_pb,
    )
    skip_elements_without_plies: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.skip_elements_without_plies"
    )
    drop_off_material = grpc_link_property("properties.drop_off_material", allowed_types=Material)
    cut_off_material = grpc_link_property("properties.cut_off_material", allowed_types=Material)
    delete_bad_elements: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.delete_bad_elements"
    )
    warping_limit: ReadWriteProperty[float, float] = grpc_data_property("properties.warping_limit")
    minimum_volume: ReadWriteProperty[float, float] = grpc_data_property(
        "properties.minimum_volume"
    )

    drop_off_settings = nested_grpc_object_property("properties.drop_off_settings", DropOffSettings)
    export_settings = nested_grpc_object_property(
        "properties.export_settings", SolidModelExportSettings
    )

    create_extrusion_guide = define_create_method(
        ExtrusionGuide,
        func_name="create_extrusion_guide",
        parent_class_name="SolidModel",
        module_name=__module__,
    )
    extrusion_guides = define_mutable_mapping(
        ExtrusionGuide, extrusion_guide_pb2_grpc.ObjectServiceStub
    )

    create_snap_to_geometry = define_create_method(
        SnapToGeometry,
        func_name="create_snap_to_geometry",
        parent_class_name="SolidModel",
        module_name=__module__,
    )
    snap_to_geometries = define_mutable_mapping(
        SnapToGeometry, snap_to_geometry_pb2_grpc.ObjectServiceStub
    )

    solid_element_sets = get_read_only_collection_property(
        SolidElementSet, solid_element_set_pb2_grpc.ObjectServiceStub
    )

    create_cut_off_geometry = define_create_method(
        CutOffGeometry,
        func_name="create_cut_off_geometry",
        parent_class_name="SolidModel",
        module_name=__module__,
    )
    cut_off_geometries = define_mutable_mapping(
        CutOffGeometry, cut_off_geometry_pb2_grpc.ObjectServiceStub
    )

    analysis_plies = get_read_only_collection_property(
        AnalysisPly, analysis_ply_pb2_grpc.ObjectServiceStub
    )
    interface_layers = get_read_only_collection_property(
        InterfaceLayer, interface_layer_pb2_grpc.ObjectServiceStub, supported_since="25.2"
    )

    elemental_data = elemental_data_property(SolidModelElementalData)
    nodal_data = nodal_data_property(SolidModelNodalData)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Generic, Protocol, TypeVar

GetValueT_co = TypeVar("GetValueT_co", covariant=True)
SetValueT_contra = TypeVar("SetValueT_contra", contravariant=True)


class ReadOnlyProperty(Generic[GetValueT_co], Protocol):
    """Interface definition for read-only properties.

    The main purpose of this protocol is to improve the type hints for
    properties which are created from helper functions.
    """

    def __get__(self, obj: object, objtype: type | None = None) -> GetValueT_co: ...


class ReadWriteProperty(Generic[GetValueT_co, SetValueT_contra], Protocol):
    """Interface definition for read-write properties.

    The main purpose of this protocol is to improve the type hints for
    properties which are created from helper functions.
    """

    def __get__(self, obj: object, objtype: type | None = None) -> GetValueT_co: ...

    def __set__(self, obj: object, value: SetValueT_contra) -> None: ...



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ._elemental_or_nodal_data import ScalarData, VectorData
from ._mesh_data import MeshData
from .analysis_ply import AnalysisPly, AnalysisPlyElementalData, AnalysisPlyNodalData
from .boolean_selection_rule import (
    BooleanSelectionRule,
    BooleanSelectionRuleElementalData,
    BooleanSelectionRuleNodalData,
)
from .butt_joint_sequence import ButtJointSequence, PrimaryPly
from .cad_component import CADComponent
from .cad_geometry import CADGeometry, TriangleMesh
from .cut_off_geometry import CutOffGeometry
from .cut_off_selection_rule import (
    CutOffSelectionRule,
    CutOffSelectionRuleElementalData,
    CutOffSelectionRuleNodalData,
)
from .cylindrical_selection_rule import (
    CylindricalSelectionRule,
    CylindricalSelectionRuleElementalData,
    CylindricalSelectionRuleNodalData,
)
from .edge_set import EdgeSet
from .element_set import ElementSet, ElementSetElementalData, ElementSetNodalData
from .enums import (
    ArrowType,
    BaseElementMaterialHandling,
    BooleanOperationType,
    CutOffGeometryOrientationType,
    CutOffMaterialHandling,
    CutOffRuleType,
    DrapingMaterialModel,
    DrapingType,
    DropOffMaterialHandling,
    DropOffType,
    EdgeSetType,
    ElementalDataType,
    ElementTechnology,
    ExtrusionGuideType,
    ExtrusionMethod,
    ExtrusionType,
    GeometricalRuleType,
    ImportedPlyDrapingType,
    ImportedPlyOffsetType,
    ImportedPlyThicknessType,
    IntersectionType,
    LookUpTable3DInterpolationAlgorithm,
    LookUpTableColumnValueType,
    MeshImportType,
    NodalDataType,
    OffsetType,
    PhysicalDimension,
    PlyCutOffType,
    PlyGeometryExportFormat,
    PlyType,
    ReinforcingBehavior,
    RosetteSelectionMethod,
    RosetteType,
    SectionCutCDBExportType,
    SectionCutType,
    SensorType,
    SnapToGeometryOrientationType,
    SolidModelExportFormat,
    SolidModelOffsetDirectionType,
    SolidModelSkinExportFormat,
    Status,
    StressStateType,
    SymmetryType,
    ThicknessFieldType,
    ThicknessType,
    UnitSystemType,
    VirtualGeometryDimension,
)
from .extrusion_guide import ExtrusionGuide
from .fabric import Fabric
from .field_definition import FieldDefinition
from .geometrical_selection_rule import (
    GeometricalSelectionRule,
    GeometricalSelectionRuleElementalData,
    GeometricalSelectionRuleNodalData,
)
from .imported_analysis_ply import ImportedAnalysisPly
from .imported_modeling_group import ImportedModelingGroup
from .imported_modeling_ply import ImportedModelingPly
from .imported_production_ply import ImportedProductionPly
from .imported_solid_model import (
    ImportedSolidModel,
    ImportedSolidModelElementalData,
    ImportedSolidModelExportSettings,
    ImportedSolidModelNodalData,
    SolidModelImportFormat,
)
from .interface_layer import InterfaceLayer
from .layup_mapping_object import LayupMappingObject, LayupMappingRosetteSelectionMethod
from .linked_selection_rule import LinkedSelectionRule
from .lookup_table_1d import LookUpTable1D
from .lookup_table_1d_column import LookUpTable1DColumn
from .lookup_table_3d import LookUpTable3D
from .lookup_table_3d_column import LookUpTable3DColumn
from .material import Material
from .model import (
    FeFormat,
    HDF5CompositeCAEImportMode,
    HDF5CompositeCAEProjectionMode,
    IgnorableEntity,
    Model,
    ModelElementalData,
    ModelNodalData,
    ShellMappingProperties,
    SolidMappingProperties,
)
from .modeling_group import ModelingGroup
from .modeling_ply import ModelingPly, ModelingPlyElementalData, ModelingPlyNodalData, TaperEdge
from .oriented_selection_set import (
    OrientedSelectionSet,
    OrientedSelectionSetElementalData,
    OrientedSelectionSetNodalData,
)
from .parallel_selection_rule import (
    ParallelSelectionRule,
    ParallelSelectionRuleElementalData,
    ParallelSelectionRuleNodalData,
)
from .production_ply import ProductionPly, ProductionPlyElementalData, ProductionPlyNodalData
from .rosette import Rosette
from .sampling_point import SamplingPoint
from .section_cut import SectionCut
from .sensor import Sensor
from .snap_to_geometry import SnapToGeometry
from .solid_element_set import (
    SolidElementSet,
    SolidElementSetElementalData,
    SolidElementSetNodalData,
)
from .solid_model import (
    DropOffSettings,
    SolidModel,
    SolidModelElementalData,
    SolidModelExportSettings,
    SolidModelNodalData,
)
from .spherical_selection_rule import (
    SphericalSelectionRule,
    SphericalSelectionRuleElementalData,
    SphericalSelectionRuleNodalData,
)
from .stackup import FabricWithAngle, Stackup
from .sublaminate import Lamina, SubLaminate
from .tube_selection_rule import (
    TubeSelectionRule,
    TubeSelectionRuleElementalData,
    TubeSelectionRuleNodalData,
)
from .utils import CoordinateTransformation
from .variable_offset_selection_rule import (
    VariableOffsetSelectionRule,
    VariableOffsetSelectionRuleElementalData,
    VariableOffsetSelectionRuleNodalData,
)
from .virtual_geometry import SubShape, VirtualGeometry

__all__ = [
    "AnalysisPly",
    "AnalysisPlyElementalData",
    "AnalysisPlyNodalData",
    "ArrowType",
    "BaseElementMaterialHandling",
    "BooleanOperationType",
    "BooleanSelectionRule",
    "BooleanSelectionRuleElementalData",
    "BooleanSelectionRuleNodalData",
    "ButtJointSequence",
    "CADComponent",
    "CADGeometry",
    "CoordinateTransformation",
    "CutOffGeometry",
    "CutOffGeometryOrientationType",
    "CutOffMaterialHandling",
    "CutOffRuleType",
    "CutOffSelectionRule",
    "CutOffSelectionRuleElementalData",
    "CutOffSelectionRuleNodalData",
    "CylindricalSelectionRule",
    "CylindricalSelectionRuleElementalData",
    "CylindricalSelectionRuleNodalData",
    "DrapingMaterialModel",
    "DrapingType",
    "DropOffMaterialHandling",
    "DropOffSettings",
    "DropOffType",
    "EdgeSet",
    "EdgeSetType",
    "ElementalDataType",
    "ElementSet",
    "ElementSetElementalData",
    "ElementSetNodalData",
    "ElementTechnology",
    "ExtrusionGuide",
    "ExtrusionGuideType",
    "ExtrusionMethod",
    "ExtrusionType",
    "Fabric",
    "FabricWithAngle",
    "FeFormat",
    "FieldDefinition",
    "FieldVariable",
    "GeometricalRuleType",
    "GeometricalSelectionRule",
    "GeometricalSelectionRuleElementalData",
    "GeometricalSelectionRuleNodalData",
    "HDF5CompositeCAEImportMode",
    "HDF5CompositeCAEProjectionMode",
    "IgnorableEntity",
    "ImportedAnalysisPly",
    "ImportedModelingGroup",
    "ImportedModelingPly",
    "ImportedPlyDrapingType",
    "ImportedPlyOffsetType",
    "ImportedPlyThicknessType",
    "ImportedProductionPly",
    "ImportedSolidModel",
    "ImportedSolidModelElementalData",
    "ImportedSolidModelExportSettings",
    "ImportedSolidModelNodalData",
    "InterfaceLayer",
    "InterpolationOptions",
    "IntersectionType",
    "Lamina",
    "LayupMappingObject",
    "LayupMappingRosetteSelectionMethod",
    "LinkedSelectionRule",
    "LookUpTable1D",
    "LookUpTable1DColumn",
    "LookUpTable3D",
    "LookUpTable3DColumn",
    "LookUpTable3DInterpolationAlgorithm",
    "LookUpTableColumnValueType",
    "Material",
    "MeshData",
    "MeshImportType",
    "Model",
    "ModelElementalData",
    "ModelingGroup",
    "ModelingPly",
    "ModelingPlyElementalData",
    "ModelingPlyNodalData",
    "ModelNodalData",
    "NodalDataType",
    "OffsetType",
    "OrientedSelectionSet",
    "OrientedSelectionSetElementalData",
    "OrientedSelectionSetNodalData",
    "ParallelSelectionRule",
    "ParallelSelectionRuleElementalData",
    "ParallelSelectionRuleNodalData",
    "PhysicalDimension",
    "PlyCutOffType",
    "PlyGeometryExportFormat",
    "PlyType",
    "PrimaryPly",
    "ProductionPly",
    "ProductionPlyElementalData",
    "ProductionPlyNodalData",
    "PuckMaterialType",
    "ReinforcingBehavior",
    "Rosette",
    "RosetteSelectionMethod",
    "RosetteType",
    "SamplingPoint",
    "ScalarData",
    "SectionCut",
    "SectionCutCDBExportType",
    "SectionCutType",
    "Sensor",
    "SensorType",
    "ShellMappingProperties",
    "SnapToGeometry",
    "SnapToGeometryOrientationType",
    "SolidElementSet",
    "SolidElementSetElementalData",
    "SolidElementSetNodalData",
    "SolidMappingProperties",
    "SolidModel",
    "SolidModelElementalData",
    "SolidModelExportFormat",
    "SolidModelExportSettings",
    "SolidModelImportFormat",
    "SolidModelNodalData",
    "SolidModelOffsetDirectionType",
    "SolidModelSkinExportFormat",
    "SphericalSelectionRule",
    "SphericalSelectionRuleElementalData",
    "SphericalSelectionRuleNodalData",
    "Stackup",
    "Status",
    "StressStateType",
    "SubLaminate",
    "SubShape",
    "SymmetryType",
    "TaperEdge",
    "ThicknessFieldType",
    "ThicknessType",
    "TriangleMesh",
    "TubeSelectionRule",
    "TubeSelectionRuleElementalData",
    "TubeSelectionRuleNodalData",
    "UnitSystemType",
    "VariableOffsetSelectionRule",
    "VariableOffsetSelectionRuleElementalData",
    "VariableOffsetSelectionRuleNodalData",
    "VectorData",
    "VirtualGeometry",
    "VirtualGeometryDimension",
]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

from ansys.api.acp.v0 import (
    lookup_table_3d_column_pb2_grpc,
    lookup_table_3d_pb2,
    lookup_table_3d_pb2_grpc,
)

from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.mapping import define_create_method, define_mutable_mapping
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_link_property,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    LookUpTable3DInterpolationAlgorithm,
    lookup_table_3d_interpolation_algorithm_from_pb,
    lookup_table_3d_interpolation_algorithm_to_pb,
    status_type_from_pb,
)
from .lookup_table_3d_column import LookUpTable3DColumn
from .object_registry import register
from .rosette import Rosette

__all__ = ["LookUpTable3D"]


@mark_grpc_properties
@register
class LookUpTable3D(CreatableTreeObject, IdTreeObject):
    """Instantiate a 3D Look-Up Table.

    The three-dimensional look-up table is defined by 3D data points. The
    locations of these data points are defined by the ``Location`` column,
    which:

    * has shape ``(N, 3)``, where ``N`` is the number of data points (rows)
    * is automatically created when the look-up table is instantiated
    * cannot be deleted or renamed

    When the length of the ``Location`` column is changed, the data of the
    other columns is automatically either truncated, or extended with ``NaN``
    values to fit.

    Parameters
    ----------
    name:
        Name of the 3D look-up table.
    use_global_coordinate_system :
        Use global coordinate system for origin and directions.
    rosette :
        Rosette used for origin and directions. Only applies if
        ``use_global_coordinate_system`` is False.
    interpolation_algorithm :
        Algorithm used to interpolate the values of the look-up table.
    use_default_search_radius :
        If ``True``, estimate the search radius used in the weightest nearest
        neighbor algorithm automatically.
    search_radius :
        Search radius used for the interpolation with the weighted nearest
        neighbor algorithm.
    num_min_neighbors :
        Minimum number of neighbors used for the interpolation when using the
        weighted nearest neighbor algorithm.
    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "lookup_tables_3d"
    _OBJECT_INFO_TYPE = lookup_table_3d_pb2.ObjectInfo
    _CREATE_REQUEST_TYPE = lookup_table_3d_pb2.CreateRequest
    _SUPPORTED_SINCE = "24.2"

    def __init__(
        self,
        *,
        name: str = "LookUpTable3D",
        use_global_coordinate_system: bool = True,
        rosette: Rosette | None = None,
        interpolation_algorithm: LookUpTable3DInterpolationAlgorithm = LookUpTable3DInterpolationAlgorithm.WEIGHTED_NEAREST_NEIGHBOR,  # noqa: E501
        use_default_search_radius: bool = True,
        search_radius: float = 0.0,
        num_min_neighbors: int = 1,
    ):
        super().__init__(name=name)
        self.use_global_coordinate_system = use_global_coordinate_system
        self.rosette = rosette
        self.interpolation_algorithm = interpolation_algorithm
        self.use_default_search_radius = use_default_search_radius
        self.search_radius = search_radius
        self.num_min_neighbors = num_min_neighbors

    def _create_stub(self) -> lookup_table_3d_pb2_grpc.ObjectServiceStub:
        return lookup_table_3d_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)

    use_global_coordinate_system: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_global_coordinate_system",
        readable_since="25.2",
        writable_since="25.2",
    )

    rosette = grpc_link_property(
        "properties.rosette", allowed_types=Rosette, readable_since="25.2", writable_since="25.2"
    )

    interpolation_algorithm = grpc_data_property(
        "properties.interpolation_algorithm",
        from_protobuf=lookup_table_3d_interpolation_algorithm_from_pb,
        to_protobuf=lookup_table_3d_interpolation_algorithm_to_pb,
    )
    use_default_search_radius: ReadWriteProperty[bool, bool] = grpc_data_property(
        "properties.use_default_search_radius"
    )
    search_radius: ReadWriteProperty[float, float] = grpc_data_property("properties.search_radius")
    num_min_neighbors: ReadWriteProperty[int, int] = grpc_data_property(
        "properties.num_min_neighbors"
    )

    create_column = define_create_method(
        LookUpTable3DColumn,
        func_name="create_column",
        parent_class_name="LookUpTable3D",
        module_name=__module__,
    )
    columns = define_mutable_mapping(
        LookUpTable3DColumn, lookup_table_3d_column_pb2_grpc.ObjectServiceStub
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import dataclasses
import typing
from typing import Any, ClassVar, Literal, cast

import numpy as np
import numpy.typing as npt
from typing_extensions import Self

if typing.TYPE_CHECKING:  # pragma: no cover
    from pyvista.core.pointset import PolyData, UnstructuredGrid

from packaging.version import Version

from ansys.acp.core._utils.array_conversions import dataarray_to_numpy, to_numpy
from ansys.api.acp.v0 import mesh_query_pb2, mesh_query_pb2_grpc

from .._utils.property_protocols import ReadOnlyProperty
from .._utils.typing_helper import StrEnum
from ._mesh_data import MeshData
from .base import TreeObject
from .enums import (
    elemental_data_type_from_pb,
    elemental_data_type_to_pb,
    nodal_data_type_from_pb,
    nodal_data_type_to_pb,
)

__all__ = [
    "ElementalData",
    "NodalData",
    "elemental_data_property",
    "nodal_data_property",
    "ScalarData",
    "VectorData",
]


@dataclasses.dataclass
class _LabelInfo:
    mesh_labels: npt.NDArray[np.int32]
    data_labels: npt.NDArray[np.int32]
    mesh_label_to_index_map: dict[np.int32, int]


def _get_labels(
    *,
    field_names: _LabelAndPyvistaFieldNames,
    labels: npt.NDArray[np.int32],
    mesh: MeshData,
) -> _LabelInfo:
    mesh_labels = getattr(mesh, field_names.LABEL_FIELD_NAME)
    mesh_label_to_index_map = {label: idx for idx, label in enumerate(mesh_labels)}
    return _LabelInfo(
        mesh_labels=mesh_labels, data_labels=labels, mesh_label_to_index_map=mesh_label_to_index_map
    )


def _expand_array(
    *,
    array: npt.NDArray[ScalarDataT],
    labels: _LabelInfo,
    culling_factor: int = 1,
) -> npt.NDArray[np.float64]:
    """Expand the array to the size of the mesh."""
    target_shape = tuple([labels.mesh_labels.size] + list(array.shape[1:]))
    target_array = np.full(target_shape, np.nan, dtype=np.float64)
    for idx, (label, value) in enumerate(zip(labels.data_labels, array)):
        if idx % culling_factor == 0:
            try:
                target_array[labels.mesh_label_to_index_map[label]] = value
            except KeyError:
                pass
    return target_array


def _get_pyvista_mesh_with_all_data(
    *,
    mesh_data_base: ElementalOrNodalDataBase,
    mesh: MeshData,
) -> UnstructuredGrid:
    pv_mesh = mesh.to_pyvista()

    mesh_data_field = getattr(
        pv_mesh, mesh_data_base._LABEL_AND_PYVISTA_FIELD_NAMES.PYVISTA_FIELD_NAME
    )
    field_labels = getattr(
        mesh_data_base, mesh_data_base._LABEL_AND_PYVISTA_FIELD_NAMES.LABEL_FIELD_NAME
    ).values
    labels = _get_labels(
        field_names=mesh_data_base._LABEL_AND_PYVISTA_FIELD_NAMES, mesh=mesh, labels=field_labels
    )

    for name in mesh_data_base._field_names():
        data = getattr(mesh_data_base, name)
        # The data can be missing if the field depends on the server version - see
        # for example the ModelElementalData._field_names method.
        if data is None:
            continue
        values = getattr(mesh_data_base, name).values
        target_array = _expand_array(array=values, labels=labels)
        mesh_data_field[name] = target_array
    return pv_mesh


def _get_mesh_with_scalar_pyvista_data(
    *,
    labels: npt.NDArray[np.int32],
    field_names: _LabelAndPyvistaFieldNames,
    mesh: MeshData,
    values: npt.NDArray[ScalarDataT],
    component_name: str,
) -> UnstructuredGrid:
    all_labels = _get_labels(field_names=field_names, labels=labels, mesh=mesh)

    pv_mesh = mesh.to_pyvista()
    mesh_data_field = getattr(pv_mesh, field_names.PYVISTA_FIELD_NAME)

    target_array = _expand_array(array=values, labels=all_labels)
    component_label = component_name
    mesh_data_field[component_label] = target_array
    return pv_mesh


def _get_pyvista_glyphs(
    *,
    labels: npt.NDArray[np.int32],
    field_names: _LabelAndPyvistaFieldNames,
    mesh: MeshData,
    values: npt.NDArray[np.float64],
    component_name: str,
    culling_factor: int = 1,
    scaling_factor: float = 1.0,
    **kwargs: Any,
) -> PolyData:
    all_labels = _get_labels(field_names=field_names, labels=labels, mesh=mesh)

    pv_mesh = mesh.to_pyvista()
    mesh_data_field = getattr(pv_mesh, field_names.PYVISTA_FIELD_NAME)

    target_array = _expand_array(array=values, labels=all_labels, culling_factor=culling_factor)
    component_label = component_name
    mesh_data_field[component_label] = target_array

    magnitude_name = f"{component_label}_magnitude"
    mesh_data_field[magnitude_name] = np.linalg.norm(target_array, axis=-1) * scaling_factor
    return pv_mesh.glyph(orient=component_label, scale=magnitude_name, **kwargs)  # type: ignore


ScalarDataT = typing.TypeVar("ScalarDataT", np.float64, np.int32)


class ScalarData(typing.Generic[ScalarDataT]):
    """Class that encapsulates scalar data."""

    def __init__(
        self,
        field_names: _LabelAndPyvistaFieldNames,
        labels: npt.NDArray[np.int32],
        values: npt.NDArray[ScalarDataT],
        component_name: str,
    ):
        self._field_names = field_names
        self._labels = labels
        self._values: npt.NDArray[ScalarDataT] = values
        self._component_name = component_name

    @property
    def values(self) -> npt.NDArray[ScalarDataT]:
        """Scalar data values as a numpy array."""
        return self._values

    @property
    def component_name(self) -> str:
        """Name of the component."""
        return self._component_name

    def get_pyvista_mesh(
        self,
        mesh: MeshData,
    ) -> UnstructuredGrid:
        """Convert the mesh data to a PyVista object.

        Parameters
        ----------
        mesh :
            The mesh to which the data is associated.
        """
        return _get_mesh_with_scalar_pyvista_data(
            labels=self._labels,
            field_names=self._field_names,
            mesh=mesh,
            values=self._values,
            component_name=self._component_name,
        )


class VectorData:
    """Class that encapsulates vector data."""

    def __init__(
        self,
        field_names: _LabelAndPyvistaFieldNames,
        labels: npt.NDArray[np.int32],
        values: npt.NDArray[np.float64],
        component_name: str,
    ):
        self._field_names = field_names
        self._labels = labels
        self._values = values
        self._component_name = component_name

    @property
    def values(self) -> npt.NDArray[np.float64]:
        """Vector data values as a numpy array."""
        return self._values

    @property
    def component_name(self) -> str:
        """Name of the component."""
        return self._component_name

    def get_pyvista_glyphs(
        self,
        *,
        mesh: MeshData,
        culling_factor: int = 1,
        scaling_factor: float = 1.0,
        **kwargs: Any,
    ) -> PolyData:
        """Get a pyvista glyph object from the vector data.

        Parameters
        ----------
        mesh :
            The mesh to which the data is associated.
        culling_factor :
            If set to a value other than ``1``, add only every n-th data
            point to the PyVista object. This is useful especially for
            vector data, where the arrows can be too dense.
        scaling_factor :
            Factor to scale the length of the arrows.
        kwargs :
            Keyword arguments passed to the PyVista object constructor.
        """
        return _get_pyvista_glyphs(
            labels=self._labels,
            field_names=self._field_names,
            mesh=mesh,
            values=self._values,
            component_name=self._component_name,
            culling_factor=culling_factor,
            scaling_factor=scaling_factor,
            **kwargs,
        )


def _check_field_type(klass: Any, field_name: str, actual_field_type: str) -> None:
    """Check that the type declared in the dataclass (klass) matches the actual type."""
    declared_field_types: typing.Sequence[str] = cast(
        typing.Sequence[str],
        [field.type for field in dataclasses.fields(klass) if field.name == field_name],
    )
    if len(declared_field_types) != 1:
        raise RuntimeError("Failed to find field in dataclass.")
    declared_field_type = declared_field_types[0].removesuffix(" | None")
    if declared_field_type != actual_field_type:
        raise RuntimeError(
            f"Declared type does not match actual data type. "
            f"Declared type: {declared_field_type}, actual type: {actual_field_type}. "
            f"Field name: {field_name}"
        )


@dataclasses.dataclass
class _LabelAndPyvistaFieldNames:
    LABEL_FIELD_NAME: str
    PYVISTA_FIELD_NAME: str


@dataclasses.dataclass
class ElementalOrNodalDataBase:
    """Base class for nodal or elemental mesh data.

    Implements the construction from a protobuf response and the conversion
    to a PyVista object.
    """

    _LABEL_AND_PYVISTA_FIELD_NAMES: ClassVar[_LabelAndPyvistaFieldNames]
    _FIELD_NAME_FROM_PB_VALUE: ClassVar[typing.Callable[[int], StrEnum]]
    _PB_VALUE_FROM_FIELD_NAME: ClassVar[typing.Callable[[StrEnum], int]]

    @classmethod
    def _field_names(cls, server_version: Version | None = None) -> list[str]:
        return [
            field.name
            for field in dataclasses.fields(cls)
            if field.name != cls._LABEL_AND_PYVISTA_FIELD_NAMES.LABEL_FIELD_NAME
        ]

    @classmethod
    def _from_pb(cls, response: mesh_query_pb2.ElementalData | mesh_query_pb2.NodalData) -> Self:
        """Construct a mesh data object from a protobuf response."""
        labels = to_numpy(response.labels)
        kwargs: dict[str, Any] = {
            cls._LABEL_AND_PYVISTA_FIELD_NAMES.LABEL_FIELD_NAME: ScalarData(
                field_names=cls._LABEL_AND_PYVISTA_FIELD_NAMES,
                labels=labels,
                values=labels,
                component_name=cls._LABEL_AND_PYVISTA_FIELD_NAMES.LABEL_FIELD_NAME,
            )
        }
        for data_type, array in zip(response.data_types, response.data_arrays):
            field_name = cls._FIELD_NAME_FROM_PB_VALUE(data_type).value
            values = cast(
                npt.NDArray[np.float64], dataarray_to_numpy(array, dtype=np.float64)
            )  # todo: handle other dtypes
            kwargs[field_name] = values
            data_wrapper: VectorData | ScalarData[np.float64]
            if len(values.shape) == 2 and values.shape[1] == 3:
                data_wrapper = VectorData(
                    field_names=cls._LABEL_AND_PYVISTA_FIELD_NAMES,
                    labels=labels,
                    values=values,
                    component_name=field_name,
                )
                _check_field_type(klass=cls, field_name=field_name, actual_field_type="VectorData")

            else:
                data_wrapper = ScalarData(
                    field_names=cls._LABEL_AND_PYVISTA_FIELD_NAMES,
                    labels=labels,
                    values=values,
                    component_name=field_name,
                )
                _check_field_type(
                    klass=cls, field_name=field_name, actual_field_type="ScalarData[np.float64]"
                )
            kwargs[field_name] = data_wrapper

        instance = cls(**kwargs)
        return instance

    def get_pyvista_mesh(
        self,
        mesh: MeshData,
    ) -> UnstructuredGrid:
        """Get a pyvista mesh with all data.

        Parameters
        ----------
        mesh :
            The mesh to which the data is associated.
        """
        return _get_pyvista_mesh_with_all_data(mesh_data_base=self, mesh=mesh)


_NODE_FIELD_NAMES = _LabelAndPyvistaFieldNames(
    LABEL_FIELD_NAME="node_labels",
    PYVISTA_FIELD_NAME="point_data",
)

_ELEMENT_FIELD_NAMES = _LabelAndPyvistaFieldNames(
    LABEL_FIELD_NAME="element_labels",
    PYVISTA_FIELD_NAME="cell_data",
)


@dataclasses.dataclass
class NodalData(ElementalOrNodalDataBase):
    """Base class for nodal data."""

    node_labels: ScalarData[np.int32]
    _LABEL_AND_PYVISTA_FIELD_NAMES = _NODE_FIELD_NAMES
    _PB_VALUE_FROM_FIELD_NAME = nodal_data_type_to_pb
    _FIELD_NAME_FROM_PB_VALUE = nodal_data_type_from_pb


@dataclasses.dataclass
class ElementalData(ElementalOrNodalDataBase):
    """Base class for elemental data."""

    element_labels: ScalarData[np.int32]
    _LABEL_AND_PYVISTA_FIELD_NAMES = _ELEMENT_FIELD_NAMES
    _PB_VALUE_FROM_FIELD_NAME = elemental_data_type_to_pb
    _FIELD_NAME_FROM_PB_VALUE = elemental_data_type_from_pb


T = typing.TypeVar("T", bound=ElementalOrNodalDataBase)


ElementalDataT = typing.TypeVar("ElementalDataT", bound=ElementalData)


def elemental_data_property(
    wrapped_cls: type[ElementalDataT],
) -> ReadOnlyProperty[ElementalDataT]:
    """Create a property to get elemental data from a tree object."""
    return _mesh_data_property_impl(
        wrapped_cls=wrapped_cls,
        request_name="GetElementalData",
        request_type=mesh_query_pb2.GetElementalDataRequest,
    )


NodalDataT = typing.TypeVar("NodalDataT", bound=NodalData)


def nodal_data_property(
    wrapped_cls: type[NodalDataT],
) -> ReadOnlyProperty[NodalDataT]:
    """Create a property to get nodal data from a tree object."""
    return _mesh_data_property_impl(
        wrapped_cls=wrapped_cls,
        request_name="GetNodalData",
        request_type=mesh_query_pb2.GetNodalDataRequest,
    )


MeshDataT = typing.TypeVar("MeshDataT", bound=ElementalOrNodalDataBase)


def _mesh_data_property_impl(
    wrapped_cls: type[MeshDataT],
    request_name: Literal["GetNodalData", "GetElementalData"],
    request_type: (
        type[mesh_query_pb2.GetNodalDataRequest] | type[mesh_query_pb2.GetElementalDataRequest]
    ),
) -> ReadOnlyProperty[MeshDataT]:
    """Create a mesh data property.

    Implementation of the mesh data property helpers ``nodal_data_property``
    and ``elemental_data_property``.
    """

    def getter(self: TreeObject) -> MeshDataT:
        if not self._is_stored:
            raise RuntimeError("Cannot get mesh data from an unstored object")
        stub = mesh_query_pb2_grpc.MeshQueryServiceStub(self._channel)
        request_func = getattr(stub, request_name)
        response = request_func(
            request=request_type(
                resource_path=self._resource_path,
                data_types=[
                    wrapped_cls._PB_VALUE_FROM_FIELD_NAME(name)  # type: ignore
                    for name in wrapped_cls._field_names(server_version=self._server_version)
                ],
            ),
        )
        return wrapped_cls._from_pb(response)

    return property(getter)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
from collections.abc import Iterable

import networkx as nx

from ._dependency_graph import _build_dependency_graph, _WalkTreeOptions
from ._tree_objects import LookUpTable1D, LookUpTable1DColumn, LookUpTable3D, LookUpTable3DColumn
from ._tree_objects._grpc_helpers.linked_object_helpers import get_linked_paths
from ._tree_objects.base import CreatableTreeObject, TreeObject
from ._utils.resource_paths import common_path, to_parts
from ._utils.typing_helper import StrEnum

__all__ = ["recursive_copy", "LinkedObjectHandling"]


class LinkedObjectHandling(StrEnum):
    """Defines options for handling linked objects when copying a tree of ACP objects."""

    KEEP = "keep"
    COPY = "copy"
    DISCARD = "discard"


def recursive_copy(
    *,
    source_objects: Iterable[CreatableTreeObject],
    parent_mapping: dict[TreeObject, TreeObject],
    linked_object_handling: LinkedObjectHandling | str = "keep",
) -> dict[CreatableTreeObject, CreatableTreeObject]:
    """Recursively copy a tree of ACP objects.

    This function copies a tree of ACP objects, starting from the given source objects.
    The copied tree includes all child objects. Linked objects (such as a Fabric linked to
    by a Modeling Ply) can optionally be included, controlled by the
    ``linked_object_handling`` argument.

    To specify where the new objects should be stored, you must provide a dictionary
    in the ``parent_mapping`` argument. The keys of the dictionary are the original
    parent objects, and the values are the new parent objects.
    Note that this mapping may need to contain parent objects that are not direct parents
    of the source objects, if another branch of the tree is included via linked objects.

    The ``parent_mapping`` argument can also include objects which are part of the
    ``source_objects`` list. In this case, the function will not create a new object for
    the parent, but will use the existing object instead.

    The function returns a ``dict`` mapping the original objects to the newly created
    objects.

    .. note::

        Only attributes supported by PyACP are copied to the new objects.

    Parameters
    ----------
    source_objects :
        The starting point of the tree to copy.
    parent_mapping :
        A list of tuples defining where the new objects are stored. Each tuple contains
        the original parent object as the first element and the new parent object as the
        second element.
    linked_object_handling :
        Defines how linked objects are handled. An example of a linked object is a Fabric
        linked to by a Modeling Ply.

        The following options are available:

        - ``"keep"``: Keep linking to the original objects, and do not
          copy them (unless they are otherwise included in the tree).
        - ``"copy"``: Copy the linked objects, and replace the links.
        - ``"discard"``: Discard object links.

        Note that when copying objects between two models, only the ``"copy"`` and
        ``"discard"`` options are valid. If you wish to use links to existing objects,
        the ``"copy"`` option can be used, specifying how links should be replaced in
        the ``parent_mapping`` argument.

    Returns
    -------
    :
        A mapping of the newly created objects. The keys are the original objects,
        and the values are the new objects.

    Examples
    --------
    To copy all Modeling Groups and associated definitions from one model to another,
    you can use the following code:

    .. code-block:: python

        import ansys.acp.core as pyacp

        model1 = ...  # loaded in some way
        model2 = ...  # loaded in some way

        pyacp.recursive_copy(
            source_objects=model1.modeling_groups.values(),
            parent_mapping={model1: model2},
            linked_object_handling="copy",
        )

    To copy all definitions from one model to another, you can use the following code:

    .. code-block:: python

        import ansys.acp.core as pyacp

        model1 = ...  # loaded in some way
        model2 = ...  # loaded in some way

        pyacp.recursive_copy(
            source_objects=[model1],
            parent_mapping={model1: model2},
            linked_object_handling="copy",
        )
    """
    # Check that the given source objects and parent mapping keys belong to the same
    # model.
    common_source_path = common_path(
        *[obj._resource_path.value for obj in list(source_objects) + list(parent_mapping.keys())]
    )
    if len(to_parts(common_source_path)) < 2:
        raise ValueError(
            "The 'source_objects' and 'parent_mapping' keys must all belong to the same model."
        )
    common_target_path = common_path(*[obj._resource_path.value for obj in parent_mapping.values()])
    if len(to_parts(common_target_path)) < 2:
        raise ValueError("The 'parent_mapping' values must all belong to the same model.")
    if linked_object_handling == LinkedObjectHandling.KEEP:
        if len(to_parts(common_path(common_source_path, common_target_path))) < 2:
            raise ValueError(
                "When using 'linked_object_handling=\"keep\"', objects cannot be copied from one model "
                "to another. The objects in 'source_objects' and 'parent_mapping' must all belong to the "
                'same model. Use \'linked_object_handling="copy"\' or linked_object_handling="discard"\' '
                "to copy objects between models."
            )

    linked_object_handling = LinkedObjectHandling(linked_object_handling)

    options = _WalkTreeOptions(
        include_children=True,
        include_linked_objects=linked_object_handling == LinkedObjectHandling.COPY,
    )
    # Build up a graph of the objects to clone. Graph edges represent a dependency:
    # - from child to parent node
    # - from source to target of a link
    graph = _build_dependency_graph(source_objects=source_objects, options=options)

    new_object_mapping: dict[CreatableTreeObject, CreatableTreeObject] = {}
    replacement_mapping = collections.ChainMap[TreeObject, TreeObject](
        new_object_mapping, parent_mapping  # type: ignore
    )

    # keep track of the new resource paths for easy replacement of linked objects
    resource_path_replacement_mapping = {
        obj._resource_path.value: new_obj._resource_path.value
        for obj, new_obj in parent_mapping.items()
    }

    # The 'topological_sort' of the graph ensures that each node is only handled
    # once its parent and linked objects are stored.
    for tree_object in reversed(list(nx.topological_sort(graph))):
        if tree_object in replacement_mapping:
            # Skip nodes which are already copied (e.g. coming from the parent_mapping)
            continue

        if isinstance(tree_object, (LookUpTable1DColumn, LookUpTable3DColumn)):
            # handled explicitly while copying the LookUpTable object
            if tree_object.name == "Location":
                continue

        new_tree_object = tree_object.clone(
            unlink=linked_object_handling == LinkedObjectHandling.DISCARD
        )

        # If the linked objects are also copied, replace them with the new objects.
        # Otherwise, we can directly store the new object.
        if linked_object_handling == LinkedObjectHandling.COPY:
            for linked_resource_path in get_linked_paths(new_tree_object._pb_object.properties):
                linked_resource_path.value = resource_path_replacement_mapping[
                    linked_resource_path.value
                ]

        try:
            new_parent = replacement_mapping[tree_object.parent]
        except KeyError as exc:
            raise KeyError(
                f"Parent object not found in 'parent_mapping' for object '{tree_object!r}'."
            ) from exc

        new_tree_object.store(parent=new_parent)

        # NOTE: if there are more type-specific fixes needed, we may want
        # to implement a more generic way to handle these.
        # Explicit fix for LookUpTable, since the Location column needs to
        # be set correctly s.t. other columns may be stored.
        if isinstance(new_tree_object, (LookUpTable1D, LookUpTable3D)):
            assert isinstance(tree_object, (LookUpTable1D, LookUpTable3D))
            new_tree_object.columns["Location"].data = tree_object.columns["Location"].data

        new_object_mapping[tree_object] = new_tree_object
        resource_path_replacement_mapping[tree_object._resource_path.value] = (
            new_tree_object._resource_path.value
        )

    return new_object_mapping



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
import dataclasses

from ansys.api.acp.v0 import solid_element_set_pb2, solid_element_set_pb2_grpc

from .._utils.array_conversions import to_tuple_from_1D_array
from .._utils.property_protocols import ReadOnlyProperty
from ._elemental_or_nodal_data import ElementalData, NodalData
from ._grpc_helpers.property_helper import grpc_data_property_read_only, mark_grpc_properties
from ._mesh_data import solid_mesh_property
from .base import IdTreeObject, ReadOnlyTreeObject
from .enums import status_type_from_pb
from .object_registry import register

__all__ = ["SolidElementSet", "SolidElementSetElementalData", "SolidElementSetNodalData"]


@dataclasses.dataclass
class SolidElementSetElementalData(ElementalData):
    """Represents elemental data for a Solid Element Set."""


@dataclasses.dataclass
class SolidElementSetNodalData(NodalData):
    """Represents nodal data for a Solid Element Set."""


@mark_grpc_properties
@register
class SolidElementSet(ReadOnlyTreeObject, IdTreeObject):
    """Instantiate a Solid Element Set.

    Parameters
    ----------
    name: str
        The name of the production ply.
    element_labels :
        Label of elements.

    """

    __slots__: Iterable[str] = tuple()

    _COLLECTION_LABEL = "solid_element_sets"
    _OBJECT_INFO_TYPE = solid_element_set_pb2.ObjectInfo
    _SUPPORTED_SINCE = "25.1"

    def _create_stub(self) -> solid_element_set_pb2_grpc.ObjectServiceStub:
        return solid_element_set_pb2_grpc.ObjectServiceStub(self._channel)

    status = grpc_data_property_read_only("properties.status", from_protobuf=status_type_from_pb)
    locked: ReadOnlyProperty[bool] = grpc_data_property_read_only("properties.locked")
    element_labels: ReadOnlyProperty[tuple[int, ...]] = grpc_data_property_read_only(
        "properties.element_labels", from_protobuf=to_tuple_from_1D_array
    )

    solid_mesh = solid_mesh_property



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable

import numpy as np
import numpy.typing as npt

from .._utils.array_conversions import to_ND_double_array_from_numpy_or_list, to_numpy
from .._utils.property_protocols import ReadWriteProperty
from ._grpc_helpers.property_helper import (
    grpc_data_property,
    grpc_data_property_read_only,
    grpc_data_setter,
    mark_grpc_properties,
)
from .base import CreatableTreeObject, IdTreeObject
from .enums import (
    LookUpTableColumnValueType,
    PhysicalDimension,
    lookup_table_column_value_type_from_pb,
    lookup_table_column_value_type_to_pb,
    physical_dimension_from_pb,
    physical_dimension_to_pb,
)

__all__ = ["LookUpTableColumnBase"]


@mark_grpc_properties
class LookUpTableColumnBase(CreatableTreeObject, IdTreeObject):
    """Base class for columns of 1D and 3D look-up tables.

    Parameters
    ----------
    value_type :
        Determines whether the column data is scalar (one entry per row) or
        directional (three entries per row).
        Note that the ``value_type`` can only be set when constructing the
        column, and is read-only afterwards.
    physical_dimension :
        Dimensionality (such as time, length, force, ...) of the column data.
    data :
        The column data. The shape of the data must match the ``value_type``
        and the length of the ``Location`` column of the parent look-up
        table.
    """

    __slots__: Iterable[str] = tuple()

    def __init__(
        self,
        *,
        name: str,
        value_type: LookUpTableColumnValueType = LookUpTableColumnValueType.SCALAR,
        physical_dimension: PhysicalDimension = PhysicalDimension.DIMENSIONLESS,
        data: npt.NDArray[np.float64] | None = None,
    ):
        super().__init__(name=name)

        # Since the value_type is settable only before the 'data' is set [1],
        # we allow it only via the constructor. To nevertheless reuse the
        # property_helper code, we simply instantiate the setter function.
        #
        # [1] In the API, the 'data' and 'value_type' need to match.
        grpc_data_setter("properties.value_type", to_protobuf=lookup_table_column_value_type_to_pb)(
            self, value_type
        )

        self.physical_dimension = physical_dimension
        if data is not None:
            self.data = data

    value_type = grpc_data_property_read_only(
        "properties.value_type", from_protobuf=lookup_table_column_value_type_from_pb
    )

    # We renamed 'dimension_type' to 'physical_dimension' compared to the API, since
    # 'dimension_type' could also be understood as the number of spatial dimensions.
    physical_dimension = grpc_data_property(
        "properties.dimension_type",
        from_protobuf=physical_dimension_from_pb,
        to_protobuf=physical_dimension_to_pb,
    )
    data: ReadWriteProperty[npt.NDArray[np.float64], npt.NDArray[np.float64]] = grpc_data_property(
        "properties.data",
        from_protobuf=to_numpy,
        to_protobuf=to_ND_double_array_from_numpy_or_list,
    )
